"""
Module with reading functionalities for atmospheric model spectra.
"""

import os
import warnings

from configparser import ConfigParser
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import dust_extinction.parameter_averages as dust_ext
import h5py
import numpy as np

from astropy import units as u
from typeguard import typechecked
from scipy.integrate import simpson
from scipy.interpolate import RegularGridInterpolator
from spectres.spectral_resampling_numba import spectres_numba

from species.core import constants
from species.core.box import (
    ColorColorBox,
    ColorMagBox,
    ModelBox,
    PhotometryBox,
    create_box,
)
from species.data.model_data.model_spectra import add_model_grid
from species.data.spec_data.spec_vega import add_vega
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_filter import ReadFilter
from species.read.read_planck import ReadPlanck
from species.util.convert_util import logg_to_mass
from species.util.dust_util import (
    convert_to_av,
    interp_lognorm,
    interp_powerlaw,
    ism_extinction,
)
from species.util.model_util import binary_to_single, check_nearest_spec, rot_int_cmj
from species.util.spec_util import smooth_spectrum


class ReadModel:
    """
    Class for reading a model spectrum from the database.
    Extinction is applied by adding the ``ext_model`` parameter
    to the ``model_param`` dictionary of any of the ``ReadModel``
    methods. The value of ``ext_model`` should be the name of
    any of the extinction models from the ``dust-extinction``
    package (see `list of available models <https://
    dust-extinction.readthedocs.io/en/latest/dust_extinction/
    choose_model.html>`_). For example, set the value to
    ``'G23'`` to use the extinction relation from `Gordon et al.
    (2023) <https://ui.adsabs.harvard.edu/abs/2023ApJ...950...86G>`_.
    When setting the ``ext_model``, the ``ext_av`` should be included
    in ``model_param`` to specify the visual extinction,
    :math:`A_V`, and optionally ``ext_rv``, to specify the
    reddening, :math:`R_V`.
    """

    @typechecked
    def __init__(
        self,
        model: str,
        wavel_range: Optional[Tuple[float, float]] = None,
        filter_name: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        model : str
            Name of the atmospheric model.
        wavel_range : tuple(float, float), None
            Wavelength range (um). Full spectrum is selected if set to
            ``None``. Not used if ``filter_name`` is not ``None``.
        filter_name : str, None
            Filter name that is used for the wavelength range. The
            ``wavel_range`` is used if set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        self.model = model

        if self.model == "bt-settl":
            warnings.warn(
                "It is recommended to use the CIFIST "
                "grid of the BT-Settl, because it is "
                "a newer version. In that case, set "
                "model='bt-settl-cifist' when using "
                "add_model of Database."
            )

        self.spectrum_interp = None
        self.wl_points = None
        self.wl_index = None

        self.filter_name = filter_name
        self.wavel_range = wavel_range

        if self.filter_name is not None:
            read_filter = ReadFilter(self.filter_name)
            self.wavel_range = read_filter.wavelength_range()
            self.mean_wavelength = read_filter.mean_wavelength()

        else:
            self.mean_wavelength = None

        if 'SPECIES_CONFIG' in os.environ:
            config_file = os.environ['SPECIES_CONFIG']
        else:
            config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]
        self.data_folder = config["species"]["data_folder"]

        self.extra_param = [
            "radius",
            "distance",
            "parallax",
            "mass",
            "log_lum",
            "log_lum_atm",
            "log_lum_disk",
            "lognorm_radius",
            "lognorm_sigma",
            "lognorm_ext",
            "ism_ext",
            "ism_red",
            "ext_model",
            "ext_av",
            "ext_rv",
            "powerlaw_max",
            "powerlaw_exp",
            "powerlaw_ext",
            "disk_teff",
            "disk_radius",
            "veil_a",
            "veil_b",
            "veil_ref",
            "vsini",
        ]

        for i in range(10):
            self.extra_param.append(f"disk_teff_{i}")
            self.extra_param.append(f"disk_radius_{i}")

        # Test if the spectra are present in the database
        hdf5_file = self.open_database()
        hdf5_file.close()

    @typechecked
    def open_database(self) -> h5py._hl.files.File:
        """
        Internal function for opening the HDF5 database.

        Returns
        -------
        h5py._hl.files.File
            The HDF5 database.
        """

        with h5py.File(self.database, "r") as hdf5_file:
            if f"models/{self.model}" in hdf5_file:
                model_found = True

            else:
                model_found = False

                warnings.warn(
                    f"The '{self.model}' model spectra are not present "
                    "in the database. Will try to add the model grid. "
                    "If this does not work (e.g. currently without an "
                    "internet connection), then please use the "
                    "'add_model' method of 'Database' to add the "
                    "grid of spectra at a later moment."
                )

        if not model_found:
            # This will not work when using multiprocessing.
            # Model spectra should be added to the database
            # before running FitModel with MPI
            with h5py.File(self.database, "a") as hdf5_file:
                add_model_grid(self.model, self.data_folder, hdf5_file)

        return h5py.File(self.database, "r")

    @typechecked
    def wavelength_points(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Internal function for extracting the wavelength points and
        indices that are used.

        Returns
        -------
        np.ndarray
            Wavelength points (um).
        np.ndarray
            Array with the size of the original wavelength grid. The
            booleans indicate if a wavelength point was used.
        """

        with self.open_database() as hdf5_file:
            wl_points = np.array(hdf5_file[f"models/{self.model}/wavelength"])

        if self.wavel_range is None:
            wl_index = np.ones(wl_points.shape[0], dtype=bool)

        else:
            wl_index = (wl_points >= self.wavel_range[0]) & (
                wl_points <= self.wavel_range[1]
            )
            index = np.where(wl_index)[0]

            # Add extra wavelength points at the boundary to make
            # sure that the wavelength range of a filter profile
            # is fully included by the model spectrum.

            # Adding 1 wavelength at both boundaries is not
            # sufficient because of the way that spectres
            # treats the edges with the resampling.

            for i in range(1, 20):
                if index[0] - i >= 0:
                    wl_index[index[0] - i] = True

                if index[-1] + i < wl_index.size:
                    wl_index[index[-1] + i] = True

        return wl_points[wl_index], wl_index

    @typechecked
    def interpolate_grid(
        self, teff_range: Optional[Tuple[float, float]] = None
    ) -> None:
        """
        Internal function for linearly interpolating the grid of
        model spectra for a requested wavelength range.

        Parameters
        ----------
        teff_range : tuple(float, float), None
            Effective temperature (K) range, (min, max) for which the
            grid will be interpolated. The full grid as stored in the
            database will be interpolated if the argument if set to
            ``None``.

        Returns
        -------
        NoneType
            None
        """

        # Get the grid points

        grid_points = self.get_points()

        # Select the required Teff points of the grid

        if "teff" in grid_points and teff_range is not None:
            teff_select = (teff_range[0] <= grid_points["teff"]) & (
                grid_points["teff"] <= teff_range[1]
            )

            # Add extra Teff points at the boundary to make sure
            # sure that the Teff prior of a fit is fully included
            # in the Teff range that is interpolated

            first_teff = np.where(teff_select)[0][0]

            if first_teff - 1 >= 0:
                teff_select[first_teff - 1] = True

            last_teff = np.where(teff_select)[0][-1]

            if last_teff + 1 < teff_select.size:
                teff_select[last_teff + 1] = True

            grid_points["teff"] = grid_points["teff"][teff_select]

        else:
            teff_select = np.ones(grid_points["teff"].size, dtype=bool)

        # Create list with grid points

        grid_points = list(grid_points.values())

        # Get the boolean array for selecting the fluxes
        # within the requested wavelength range

        if self.wl_index is None:
            self.wl_points, self.wl_index = self.wavelength_points()

        # Open de HDF5 database and read the model fluxes

        with self.open_database() as hdf5_file:
            grid_flux = np.array(hdf5_file[f"models/{self.model}/flux"])
            grid_flux = grid_flux[..., self.wl_index]
            grid_flux = grid_flux[teff_select, ...]

        # Interpolate the grid of model spectra

        self.spectrum_interp = RegularGridInterpolator(
            grid_points,
            grid_flux,
            method="linear",
            bounds_error=True,
            fill_value=np.nan,
        )

    @typechecked
    def apply_lognorm_ext(
        self,
        wavelength: np.ndarray,
        flux: np.ndarray,
        lognorm_radius: float,
        lognorm_sigma: float,
        lognorm_ext: float,
    ) -> np.ndarray:
        """
        Internal function for applying extinction by dust to a spectrum.

        wavelength : np.ndarray
            Wavelengths (um) of the spectrum.
        flux : np.ndarray
            Fluxes (W m-2 um-1) of the spectrum.
        lognorm_radius : float
            Logarithm (base 10) of the mean geometric radius (um)
            of the log-normal size distribution.
        lognorm_sigma : float
            Geometric standard deviation (dimensionless) of the
            log-normal size distribution.
        lognorm_ext : float
            The extinction (mag) in the V band.

        Returns
        -------
        np.ndarray
            Fluxes (W m-2 um-1) with the extinction applied.
        """

        # Interpolate cross sections as function of wavelength,
        # geometric radius, and geometric standard deviation

        dust_interp, _, _ = interp_lognorm(verbose=False)
        dust_wavel = dust_interp.grid[0]

        if wavelength[0] < dust_wavel[0]:
            raise ValueError(
                f"The shortest wavelength ({wavelength[0]:.2e} um) "
                "for which the spectrum will be calculated is smaller "
                f"than the shortest wavelength ({dust_wavel[0]:.2e} "
                "um) of the grid with dust cross sections."
            )

        if wavelength[-1] > dust_wavel[-1]:
            raise ValueError(
                f"The longest wavelength ({wavelength[-1]:.2e} um) "
                "for which the spectrum  will be calculated is "
                "larger than the longest wavelength "
                f"({dust_wavel[-1]:.2e} um) of the grid with dust "
                "cross sections."
            )

        # For each radius-sigma pair, cross sections are normalized
        # by the integrated cross section in the V-band

        cross_sections = dust_interp((wavelength, 10.0**lognorm_radius, lognorm_sigma))

        return flux * np.exp(-lognorm_ext * cross_sections)

    @typechecked
    def apply_powerlaw_ext(
        self,
        wavelength: np.ndarray,
        flux: np.ndarray,
        powerlaw_max: float,
        powerlaw_exp: float,
        powerlaw_ext: float,
    ) -> np.ndarray:
        """
        Internal function for applying extinction by dust to a
        spectrum.

        wavelength : np.ndarray
            Wavelengths (um) of the spectrum.
        flux : np.ndarray
            Fluxes (W m-2 um-1) of the spectrum.
        powerlaw_max : float
            Logarithm (base 10) of the maximum radius (um)
            of the power-law size distribution.
        powerlaw_exp : float
            Exponent of the power-law size distribution.
        powerlaw_ext : float
            The extinction (mag) in the V band.

        Returns
        -------
        np.ndarray
            Fluxes (W m-2 um-1) with the extinction applied.
        """

        # Interpolate cross sections as function of wavelength,
        # geometric radius, and geometric standard deviation

        dust_interp, _, _ = interp_powerlaw(verbose=False)
        dust_wavel = dust_interp.grid[0]

        if wavelength[0] < dust_wavel[0]:
            raise ValueError(
                f"The shortest wavelength ({wavelength[0]:.2e} um) for which the "
                f"spectrum will be calculated is smaller than the shortest "
                f"wavelength ({dust_wavel[0]:.2e} um) of the grid with dust cross "
                f"sections."
            )

        if wavelength[-1] > dust_wavel[-1]:
            raise ValueError(
                f"The longest wavelength ({wavelength[-1]:.2e} um) for which the "
                f"spectrum  will be calculated is larger than the longest wavelength "
                f"({dust_wavel[-1]:.2e} um) of the grid with dust cross sections."
            )

        # For each radius-sigma pair, cross sections are normalized
        # by the integrated cross section in the V-band

        cross_sections = dust_interp((wavelength, 10.0**powerlaw_max, powerlaw_exp))

        return flux * np.exp(-powerlaw_ext * cross_sections)

    @staticmethod
    @typechecked
    def apply_ext_ism(
        wavelengths: np.ndarray, flux: np.ndarray, v_band_ext: float, v_band_red: float
    ) -> np.ndarray:
        """
        Internal function for applying ISM extinction to a spectrum.

        wavelengths : np.ndarray
            Wavelengths (um) of the spectrum.
        flux : np.ndarray
            Fluxes (W m-2 um-1) of the spectrum.
        v_band_ext : float
            Extinction (mag) in the V band.
        v_band_red : float
            Reddening in the V band.

        Returns
        -------
        np.ndarray
            Fluxes (W m-2 um-1) with the extinction applied.
        """

        ext_mag = ism_extinction(v_band_ext, v_band_red, wavelengths)

        return flux * 10.0 ** (-0.4 * ext_mag)

    @typechecked
    def get_model(
        self,
        model_param: Dict[str, float],
        spec_res: Optional[float] = None,
        wavel_resample: Optional[np.ndarray] = None,
        magnitude: bool = False,
        ext_filter: Optional[str] = None,
        **kwargs,
    ) -> ModelBox:
        """
        Function for extracting a model spectrum by linearly
        interpolating the model grid.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The grid
            boundaries of the spectra in the database can be obtained
            with
            :func:`~species.read.read_model.ReadModel.get_bounds()`.
        spec_res : float, None
            Spectral resolution that is used for smoothing the spectrum
            with a Gaussian kernel. No smoothing is applied if the
            argument is set to ``None``.
        wavel_resample : np.ndarray, None
            Wavelength points (um) to which the spectrum is resampled.
            Optional smoothing with ``spec_res`` is applied for
            resampling with ``wavel_resample``. The wavelength points
            as stored in the database are used if the argument is set
            to ``None``.
        magnitude : bool
            Normalize the spectrum with a flux calibrated spectrum of
            Vega and return the magnitude instead of flux density.
        ext_filter : str, None
            Filter that is associated with the (optional) extinction
            parameter, ``ism_ext``. When the argument of ``ext_filter``
            is set to ``None``, the extinction is defined in the visual
            as usual (i.e. :math:`A_V`). By providing a filter name
            from the `SVO Filter Profile Service <http://svo2.cab.
            inta-csic.es/svo/theory/fps/>`_ as argument then the
            extinction ``ism_ext`` is defined in that filter instead
            of the $V$ band.

        Returns
        -------
        species.core.box.ModelBox
            Box with the model spectrum.
        """

        if "smooth" in kwargs:
            warnings.warn(
                "The 'smooth' parameter has been "
                "deprecated. Please set only the "
                "'spec_res' argument, which can be set "
                "to None for not applying a smoothing.",
                DeprecationWarning,
            )

            if not kwargs["smooth"] and spec_res is not None:
                spec_res = None

        # Check nearest grid points

        check_nearest_spec(self.model, model_param)

        # Get grid boundaries

        grid_bounds = self.get_bounds()

        # Check if all parameters are present and within the grid boundaries

        param_list = self.get_parameters()

        for key in param_list:
            if key not in model_param.keys():
                raise ValueError(
                    f"The '{key}' parameter is required by '{self.model}'. "
                    f"The mandatory parameters are {param_list}."
                )

            if model_param[key] < grid_bounds[key][0]:
                raise ValueError(
                    f"The input value of '{key}' is smaller than the lower "
                    f"boundary of the model grid ({model_param[key]} < "
                    f"{grid_bounds[key][0]})."
                )

            if model_param[key] > grid_bounds[key][1]:
                raise ValueError(
                    f"The input value of '{key}' is larger than the upper "
                    f"boundary of the model grid ({model_param[key]} > "
                    f"{grid_bounds[key][1]})."
                )

        # Print a warning if redundant parameters are included in the dictionary

        ignore_param = []

        param_list = self.get_parameters()

        for key in model_param.keys():
            if (
                key not in param_list
                and key not in self.extra_param
                and not key.startswith("phot_ext_")
            ):
                warnings.warn(
                    f"The '{key}' parameter is not required by "
                    f"'{self.model}' so the parameter will be "
                    f"ignored. The mandatory parameters are "
                    f"{param_list}."
                )

                ignore_param.append(key)

        # Interpolate the model grid

        if self.spectrum_interp is None:
            self.interpolate_grid()

        # Set the wavelength range

        if self.wavel_range is None:
            wl_points = self.get_wavelengths()
            self.wavel_range = (wl_points[0], wl_points[-1])

        # Create a list with the parameter values

        check_param = [
            "teff",
            "logg",
            "feh",
            "c_o_ratio",
            "fsed",
            "log_kzz",
            "ad_index",
            "log_co_iso",
        ]

        parameters = []
        for item in check_param:
            if item in model_param and item not in ignore_param:
                parameters.append(model_param[item])

        # Check if the ext_filter should be adjusted
        # to the name that is extracted from the
        # phot_ext_{ext_filter} parameter

        if ext_filter is None:
            for param_item in model_param:
                if param_item.startswith("phot_ext_"):
                    ext_filter = param_item[9:]

        # Interpolate the spectrum from the grid

        flux = self.spectrum_interp(parameters)[0]

        # Add the radius to the parameter dictionary if the mass if given

        if "mass" in model_param and "radius" not in model_param:
            mass = 1e3 * model_param["mass"] * constants.M_JUP  # (g)
            radius = np.sqrt(
                1e3 * constants.GRAVITY * mass / (10.0 ** model_param["logg"])
            )  # (cm)
            model_param["radius"] = 1e-2 * radius / constants.R_JUP  # (Rjup)

        # Apply (radius/distance)^2 scaling

        if "radius" in model_param and "parallax" in model_param:
            scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                1e3 * constants.PARSEC / model_param["parallax"]
            ) ** 2

            flux *= scaling

        elif "radius" in model_param and "distance" in model_param:
            scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                model_param["distance"] * constants.PARSEC
            ) ** 2

            flux *= scaling

        elif "flux_scaling" in model_param:
            flux *= model_param["flux_scaling"]

        elif "log_flux_scaling" in model_param:
            flux *= 10.0 ** model_param["log_flux_scaling"]

        # Add optional offset to the flux

        if "flux_offset" in model_param:
            flux += model_param["flux_offset"]

        # Add blackbody disk component to the spectrum

        n_disk = 0

        if "disk_teff" in model_param and "disk_radius" in model_param:
            n_disk = 1

        else:
            for disk_idx in range(100):
                if (
                    f"disk_teff_{disk_idx}" in model_param
                    and f"disk_radius_{disk_idx}" in model_param
                ):
                    n_disk += 1
                else:
                    break

        if n_disk == 1:
            readplanck = ReadPlanck(
                (0.9 * self.wavel_range[0], 1.1 * self.wavel_range[-1])
            )

            disk_param = {
                "teff": model_param["disk_teff"],
                "radius": model_param["disk_radius"],
            }

            if "parallax" in model_param:
                disk_param["parallax"] = model_param["parallax"]

            elif "distance" in model_param:
                disk_param["distance"] = model_param["distance"]

            planck_box = readplanck.get_spectrum(disk_param, spec_res=spec_res)

            flux += spectres_numba(
                self.wl_points,
                planck_box.wavelength,
                planck_box.flux,
                spec_errs=None,
                fill=np.nan,
                verbose=True,
            )

        elif n_disk > 1:
            readplanck = ReadPlanck(
                (0.9 * self.wavel_range[0], 1.1 * self.wavel_range[-1])
            )

            for disk_idx in range(n_disk):
                disk_param = {
                    "teff": model_param[f"disk_teff_{disk_idx}"],
                    "radius": model_param[f"disk_radius_{disk_idx}"],
                }

                if "parallax" in model_param:
                    disk_param["parallax"] = model_param["parallax"]

                elif "distance" in model_param:
                    disk_param["distance"] = model_param["distance"]

                planck_box = readplanck.get_spectrum(disk_param, spec_res=spec_res)

                flux += spectres_numba(
                    self.wl_points,
                    planck_box.wavelength,
                    planck_box.flux,
                    spec_errs=None,
                    fill=np.nan,
                    verbose=True,
                )

        # Create ModelBox with the spectrum

        model_box = create_box(
            boxtype="model",
            model=self.model,
            wavelength=self.wl_points,
            flux=flux,
            parameters=model_param,
            quantity="flux",
            spec_res=spec_res,
        )

        # Apply rotational broadening vsin(i) in km/s

        if "vsini" in model_param:
            model_box.flux = rot_int_cmj(
                wavel=model_box.wavelength,
                flux=model_box.flux,
                vsini=model_param["vsini"],
                eps=0.0,
            )

        # Apply veiling

        if (
            "veil_a" in model_param
            and "veil_b" in model_param
            and "veil_ref" in model_param
        ):
            lambda_ref = 0.5  # (um)

            veil_flux = model_param["veil_ref"] + model_param["veil_b"] * (
                model_box.wavelength - lambda_ref
            )

            model_box.flux = model_param["veil_a"] * model_box.flux + veil_flux

        # Apply extinction

        if (
            "lognorm_radius" in model_param
            and "lognorm_sigma" in model_param
            and "lognorm_ext" in model_param
        ):
            model_box.flux = self.apply_lognorm_ext(
                model_box.wavelength,
                model_box.flux,
                model_param["lognorm_radius"],
                model_param["lognorm_sigma"],
                model_param["lognorm_ext"],
            )

        if (
            "powerlaw_max" in model_param
            and "powerlaw_exp" in model_param
            and "powerlaw_ext" in model_param
        ):
            model_box.flux = self.apply_powerlaw_ext(
                model_box.wavelength,
                model_box.flux,
                model_param["powerlaw_max"],
                model_param["powerlaw_exp"],
                model_param["powerlaw_ext"],
            )

        if "ism_ext" in model_param or ext_filter is not None:
            ism_reddening = model_param.get("ism_red", 3.1)

            if ext_filter is not None:
                ism_ext_av = convert_to_av(
                    filter_name=ext_filter,
                    filter_ext=model_param[f"phot_ext_{ext_filter}"],
                    v_band_red=ism_reddening,
                )

            else:
                ism_ext_av = model_param["ism_ext"]

            model_box.flux = self.apply_ext_ism(
                model_box.wavelength,
                model_box.flux,
                ism_ext_av,
                ism_reddening,
            )

        if "ext_av" in model_param:
            if "ext_model" in model_param:
                ext_model = getattr(dust_ext, model_param["ext_model"])()

                if "ext_rv" in model_param:
                    ext_model.Rv = model_param["ext_rv"]

                # Wavelength range (um) for which the extinction is defined
                ext_wavel = (1.0 / ext_model.x_range[1], 1.0 / ext_model.x_range[0])

                if (
                    model_box.wavelength[0] < ext_wavel[0]
                    or model_box.wavelength[-1] > ext_wavel[1]
                ):
                    warnings.warn(
                        "The wavelength range of the model spectrum "
                        f"({model_box.wavelength[0]:.3f}-"
                        f"{model_box.wavelength[-1]:.3f} um) "
                        "does not fully lie within the available "
                        "wavelength range of the extinction model "
                        f"({ext_wavel[0]:.3f}-{ext_wavel[1]:.3f} um). "
                        "The extinction will therefore not be applied "
                        "to fluxes of which the wavelength lies "
                        "outside the range of the extinction model."
                    )

                wavel_select = (model_box.wavelength > ext_wavel[0]) & (
                    model_box.wavelength < ext_wavel[1]
                )

                model_box.flux[wavel_select] *= ext_model.extinguish(
                    model_box.wavelength[wavel_select] * u.micron,
                    Av=model_param["ext_av"],
                )

            else:
                warnings.warn(
                    "The 'ext_av' parameter is included in the "
                    "'model_param' dictionary but the 'ext_model' "
                    "parameter is missing. Therefore, the 'ext_av' "
                    "parameter is ignored and no extinction is "
                    "applied to the spectrum."
                )

        # Apply radial velocity shift to the wavelengths

        if "rad_vel" in model_param:
            # Wavelength shift in um
            # rad_vel in km s-1 and constants.LIGHT in m s-1

            wavel_shift = (
                model_box.wavelength * 1e3 * model_param["rad_vel"] / constants.LIGHT
            )

            # Resampling will introduce a few NaNs at the edge of the
            # flux array. Resampling is needed because shifting the
            # wavelength array does not work when combining two spectra
            # of a binary system of which the two stars have different RVs.

            model_box.flux = spectres_numba(
                model_box.wavelength,
                model_box.wavelength + wavel_shift,
                model_box.flux,
                spec_errs=None,
                fill=np.nan,
                verbose=False,
            )

        # Smooth the spectrum

        if spec_res is not None:
            model_box.flux = smooth_spectrum(
                model_box.wavelength, model_box.flux, spec_res
            )

        # Resample the spectrum

        if wavel_resample is not None:
            model_box.flux = spectres_numba(
                wavel_resample,
                model_box.wavelength,
                model_box.flux,
                spec_errs=None,
                fill=np.nan,
                verbose=True,
            )

            model_box.wavelength = wavel_resample

        # Convert flux to magnitude

        if magnitude:
            with h5py.File(self.database, "r") as hdf5_file:
                # Check if the Vega spectrum is found in
                # 'r' mode because the 'a' mode is not
                # possible when using multiprocessing
                vega_found = "spectra/calibration/vega" in hdf5_file

            if not vega_found:
                with h5py.File(self.database, "a") as hdf5_file:
                    add_vega(self.data_folder, hdf5_file)

            with h5py.File(self.database, "r") as hdf5_file:
                vega_spec = np.array(hdf5_file["spectra/calibration/vega"])

            flux_vega = spectres_numba(
                model_box.wavelength,
                vega_spec[0,],
                vega_spec[1,],
                spec_errs=None,
                fill=np.nan,
                verbose=True,
            )

            model_box.flux = -2.5 * np.log10(model_box.flux / flux_vega)
            model_box.quantity = "magnitude"

        # Check if the contains NaNs

        if np.isnan(np.sum(model_box.flux)):
            warnings.warn(
                "The resampled spectrum contains "
                f"{np.sum(np.isnan(model_box.flux))} "
                "NaNs, probably because the original "
                "wavelength range does not fully encompass "
                "the new wavelength range. This happened with "
                f"the following parameters: {model_param}."
            )

        # Add the luminosity to the parameter dictionary

        lum_total = 0.0

        if "radius" in model_box.parameters:
            lum_atm = (
                4.0
                * np.pi
                * (model_box.parameters["radius"] * constants.R_JUP) ** 2
                * constants.SIGMA_SB
                * model_box.parameters["teff"] ** 4.0
                / constants.L_SUN
            )  # (Lsun)

            lum_total += lum_atm
            model_box.parameters["log_lum_atm"] = np.log10(lum_atm)

        # Add the blackbody disk components to the luminosity

        if n_disk == 1:
            lum_disk = (
                4.0
                * np.pi
                * (model_box.parameters["disk_radius"] * constants.R_JUP) ** 2
                * constants.SIGMA_SB
                * model_box.parameters["disk_teff"] ** 4.0
                / constants.L_SUN
            )  # (Lsun)

            lum_total += lum_disk
            model_box.parameters["log_lum_disk"] = np.log10(lum_disk)

        elif n_disk > 1:
            for disk_idx in range(n_disk):
                lum_disk = (
                    4.0
                    * np.pi
                    * (
                        model_box.parameters[f"disk_radius_{disk_idx}"]
                        * constants.R_JUP
                    )
                    ** 2
                    * constants.SIGMA_SB
                    * model_box.parameters[f"disk_teff_{disk_idx}"] ** 4.0
                    / constants.L_SUN
                )  # (Lsun)

                lum_total += lum_disk
                model_box.parameters[f"log_lum_disk_{disk_idx}"] = np.log10(lum_disk)

        if lum_total > 0.0:
            model_box.parameters["log_lum"] = np.log10(lum_total)

        # Add the planet mass to the parameter dictionary

        if "radius" in model_param and "logg" in model_param:
            model_param["mass"] = logg_to_mass(
                model_param["logg"], model_param["radius"]
            )

        return model_box

    @typechecked
    def get_data(
        self,
        model_param: Dict[str, float],
        spec_res: Optional[float] = None,
        wavel_resample: Optional[np.ndarray] = None,
        ext_filter: Optional[str] = None,
    ) -> ModelBox:
        """
        Function for selecting a model spectrum (without interpolation)
        for a set of parameter values that coincide with the grid
        points. The stored grid points can be inspected with
        :func:`~species.read.read_model.ReadModel.get_points`.

        Parameters
        ----------
        model_param : dict
            Model parameters and values. Only discrete values from the
            original grid are possible. Else, the nearest grid values
            are selected.
        spec_res : float, None
            Spectral resolution that is used for smoothing the spectrum
            with a Gaussian kernel. No smoothing is applied to the
            spectrum if the argument is set to ``None``.
        wavel_resample : np.ndarray, None
            Wavelength points (um) to which the spectrum will be
            resampled. In that case, ``spec_res`` can still be used for
            smoothing the spectrum with a Gaussian kernel. The original
            wavelength points are used if the argument is set to
            ``None``.
        ext_filter : str, None
            Filter that is associated with the (optional) extinction
            parameter, ``ism_ext``. When the argument of ``ext_filter``
            is set to ``None``, the extinction is defined in the visual
            as usual (i.e. :math:`A_V`). By providing a filter name
            from the `SVO Filter Profile Service <http://svo2.cab.
            inta-csic.es/svo/theory/fps/>`_ as argument then the
            extinction ``ism_ext`` is defined in that filter instead
            of the $V$ band.

        Returns
        -------
        species.core.box.ModelBox
            Box with the model spectrum.
        """

        # Check if all parameters are present

        param_list = self.get_parameters()

        for key in param_list:
            if key not in model_param.keys():
                raise ValueError(
                    f"The '{key}' parameter is required by '{self.model}'. "
                    f"The mandatory parameters are {param_list}."
                )

        # Print a warning if redundant parameters are included in the dictionary

        ignore_param = []

        for key in model_param.keys():
            if key not in param_list and key not in self.extra_param:
                warnings.warn(
                    f"The '{key}' parameter is not required by "
                    f"'{self.model}' so the parameter will be "
                    f"ignored. The mandatory parameters are "
                    f"{param_list}."
                )

                ignore_param.append(key)

        # Get wavelength points for wavelength range of
        # a filter in case filter_name was set or a
        # spectrum in case wavel_range was set

        self.wl_points, self.wl_index = self.wavelength_points()

        # Model parameters to check

        check_param = [
            "teff",
            "logg",
            "feh",
            "c_o_ratio",
            "fsed",
            "log_kzz",
            "ad_index",
            "log_co_iso",
        ]

        # Create lists with the parameter names and values

        param_key = []
        param_val = []

        for param_item in check_param:
            if param_item in model_param and param_item not in ignore_param:
                param_key.append(param_item)
                param_val.append(model_param[param_item])

        # Check if the ext_filter should be adjusted
        # to the name that is extracted from the
        # phot_ext_{ext_filter} parameter

        if ext_filter is None:
            for param_item in model_param:
                if param_item.startswith("phot_ext_"):
                    ext_filter = param_item[9:]

        # Open de HDF5 database

        with self.open_database() as hdf5_file:
            # Read the grid of fluxes from the database

            flux = np.array(hdf5_file[f"models/{self.model}/flux"])
            flux = flux[..., self.wl_index]

            # Find the indices of the grid points for which the spectrum will be extracted

            indices = []

            for i, item in enumerate(param_key):
                data = np.array(hdf5_file[f"models/{self.model}/{item}"])
                data_index = np.argwhere(
                    np.round(data, 4) == np.round(model_param[item], 4)
                )

                if len(data_index) == 0:
                    raise ValueError(
                        f"The parameter {item}={param_val[i]} is not found."
                    )

                data_index = data_index[0]

                indices.append(data_index[0])

        # Extract the spectrum at the requested grid point

        flux = flux[tuple(indices)]

        # Apply (radius/distance)^2 scaling

        if "radius" in model_param and "parallax" in model_param:
            scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                1e3 * constants.PARSEC / model_param["parallax"]
            ) ** 2

            flux *= scaling

        elif "radius" in model_param and "distance" in model_param:
            scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                model_param["distance"] * constants.PARSEC
            ) ** 2

            flux *= scaling

        elif "flux_scaling" in model_param:
            flux *= model_param["flux_scaling"]

        elif "log_flux_scaling" in model_param:
            flux *= 10.0 ** model_param["log_flux_scaling"]

        # Add optional offset to the flux

        if "flux_offset" in model_param:
            flux += model_param["flux_offset"]

        # Add blackbody disk component to the spectrum

        n_disk = 0

        if "disk_teff" in model_param and "disk_radius" in model_param:
            n_disk = 1

        else:
            for disk_idx in range(100):
                if "disk_teff_{disk_idx}" in model_param:
                    n_disk += 1
                else:
                    break

        if n_disk == 1:
            readplanck = ReadPlanck(
                (0.9 * self.wavel_range[0], 1.1 * self.wavel_range[-1])
            )

            disk_param = {
                "teff": model_param["disk_teff"],
                "radius": model_param["disk_radius"],
            }

            if "parallax" in model_param:
                disk_param["parallax"] = model_param["parallax"]

            elif "distance" in model_param:
                disk_param["distance"] = model_param["distance"]

            planck_box = readplanck.get_spectrum(disk_param, spec_res=spec_res)

            flux += spectres_numba(
                self.wl_points,
                planck_box.wavelength,
                planck_box.flux,
                spec_errs=None,
                fill=np.nan,
                verbose=True,
            )

        elif n_disk > 1:
            readplanck = ReadPlanck(
                (0.9 * self.wavel_range[0], 1.1 * self.wavel_range[-1])
            )

            for disk_idx in range(n_disk):
                disk_param = {
                    "teff": model_param[f"disk_teff_{disk_idx}"],
                    "radius": model_param[f"disk_radius_{disk_idx}"],
                }

                if "parallax" in model_param:
                    disk_param["parallax"] = model_param["parallax"]

                elif "distance" in model_param:
                    disk_param["distance"] = model_param["distance"]

                planck_box = readplanck.get_spectrum(disk_param, spec_res=spec_res)

                flux += spectres_numba(
                    self.wl_points,
                    planck_box.wavelength,
                    planck_box.flux,
                    spec_errs=None,
                    fill=np.nan,
                    verbose=True,
                )

        # Create ModelBox with the spectrum

        model_box = create_box(
            boxtype="model",
            model=self.model,
            wavelength=self.wl_points,
            flux=flux,
            parameters=model_param,
            quantity="flux",
            spec_res=spec_res,
        )

        # Apply rotational broadening vsin(i) in km/s

        if "vsini" in model_param:
            model_box.flux = rot_int_cmj(
                wavel=model_box.wavelength,
                flux=model_box.flux,
                vsini=model_param["vsini"],
                eps=0.0,
            )

        # Apply veiling

        if (
            "veil_a" in model_param
            and "veil_b" in model_param
            and "veil_ref" in model_param
        ):
            lambda_ref = 0.5  # (um)

            veil_flux = model_param["veil_ref"] + model_param["veil_b"] * (
                model_box.wavelength - lambda_ref
            )

            model_box.flux = model_param["veil_a"] * model_box.flux + veil_flux

        # Apply extinction

        if (
            "lognorm_radius" in model_param
            and "lognorm_sigma" in model_param
            and "lognorm_ext" in model_param
        ):
            raise NotImplementedError(
                "Log-normal dust is nog yet implemented for get_data."
            )

        if (
            "powerlaw_max" in model_param
            and "powerlaw_exp" in model_param
            and "powerlaw_ext" in model_param
        ):
            raise NotImplementedError(
                "Power-law dust is nog yet implemented for get_data."
            )

        if "ism_ext" in model_param or ext_filter is not None:
            ism_reddening = model_param.get("ism_red", 3.1)

            if ext_filter is not None:
                ism_ext_av = convert_to_av(
                    filter_name=ext_filter,
                    filter_ext=model_param[f"phot_ext_{ext_filter}"],
                    v_band_red=ism_reddening,
                )

            else:
                ism_ext_av = model_param["ism_ext"]

            model_box.flux = self.apply_ext_ism(
                model_box.wavelength,
                model_box.flux,
                ism_ext_av,
                ism_reddening,
            )

        if "ext_av" in model_param:
            if "ext_model" in model_param:
                ext_model = getattr(dust_ext, model_param["ext_model"])()

                if "ext_rv" in model_param:
                    ext_model.Rv = model_param["ext_rv"]

                # Wavelength range (um) for which the extinction is defined
                ext_wavel = (1.0 / ext_model.x_range[1], 1.0 / ext_model.x_range[0])

                if (
                    model_box.wavelength[0] < ext_wavel[0]
                    or model_box.wavelength[-1] > ext_wavel[1]
                ):
                    warnings.warn(
                        "The wavelength range of the model spectrum "
                        f"({model_box.wavelength[0]:.3f}-"
                        f"{model_box.wavelength[-1]:.3f} um) "
                        "does not fully lie within the available "
                        "wavelength range of the extinction model "
                        f"({ext_wavel[0]:.3f}-{ext_wavel[1]:.3f} um). "
                        "The extinction will therefore not be applied "
                        "to fluxes of which the wavelength lies "
                        "outside the range of the extinction model."
                    )

                wavel_select = (model_box.wavelength > ext_wavel[0]) & (
                    model_box.wavelength < ext_wavel[1]
                )

                model_box.flux[wavel_select] *= ext_model.extinguish(
                    model_box.wavelength[wavel_select] * u.micron,
                    Av=model_param["ext_av"],
                )

            else:
                warnings.warn(
                    "The 'ext_av' parameter is included in the "
                    "'model_param' dictionary but the 'ext_model' "
                    "parameter is missing. Therefore, the 'ext_av' "
                    "parameter is ignored and no extinction is "
                    "applied to the spectrum."
                )

        # Apply radial velocity shift to the wavelengths

        if "rad_vel" in model_param:
            # Wavelength shift in um
            # rad_vel in km s-1 and constants.LIGHT in m s-1

            wavel_shift = (
                model_box.wavelength * 1e3 * model_param["rad_vel"] / constants.LIGHT
            )

            # Resampling will introduce a few NaNs at the edge of the
            # flux array. Resampling is needed because shifting the
            # wavelength array does not work when combining two spectra
            # of a binary system of which the two stars have different RVs.

            model_box.flux = spectres_numba(
                model_box.wavelength,
                model_box.wavelength + wavel_shift,
                model_box.flux,
                spec_errs=None,
                fill=np.nan,
                verbose=False,
            )

        # Smooth the spectrum

        if spec_res is not None:
            model_box.flux = smooth_spectrum(
                model_box.wavelength, model_box.flux, spec_res
            )

        # Resample the spectrum

        if wavel_resample is not None:
            model_box.flux = spectres_numba(
                wavel_resample,
                model_box.wavelength,
                model_box.flux,
                spec_errs=None,
                fill=np.nan,
                verbose=True,
            )

            model_box.wavelength = wavel_resample

        # Add the luminosity to the parameter dictionary

        lum_total = 0.0

        if "radius" in model_box.parameters:
            lum_atm = (
                4.0
                * np.pi
                * (model_box.parameters["radius"] * constants.R_JUP) ** 2
                * constants.SIGMA_SB
                * model_box.parameters["teff"] ** 4.0
                / constants.L_SUN
            )  # (Lsun)

            lum_total += lum_atm
            model_box.parameters["log_lum_atm"] = np.log10(lum_atm)

        # Add the blackbody disk components to the luminosity

        if n_disk == 1:
            lum_disk = (
                4.0
                * np.pi
                * (model_box.parameters["disk_radius"] * constants.R_JUP) ** 2
                * constants.SIGMA_SB
                * model_box.parameters["disk_teff"] ** 4.0
                / constants.L_SUN
            )  # (Lsun)

            lum_total += lum_disk
            model_box.parameters["log_lum_disk"] = np.log10(lum_disk)

        elif n_disk > 1:
            for disk_idx in range(n_disk):
                lum_disk = (
                    4.0
                    * np.pi
                    * (
                        model_box.parameters[f"disk_radius_{disk_idx}"]
                        * constants.R_JUP
                    )
                    ** 2
                    * constants.SIGMA_SB
                    * model_box.parameters["disk_teff"] ** 4.0
                    / constants.L_SUN
                )  # (Lsun)

                lum_total += lum_disk
                model_box.parameters[f"log_lum_disk_{disk_idx}"] = np.log10(lum_disk)

        if lum_total > 0.0:
            model_box.parameters["log_lum"] = np.log10(lum_total)

        # Add the planet mass to the parameter dictionary

        if "radius" in model_param and "logg" in model_param:
            model_param["mass"] = logg_to_mass(
                model_param["logg"], model_param["radius"]
            )

        return model_box

    @typechecked
    def get_flux(
        self, model_param: Dict[str, float], synphot=None, return_box: bool = False
    ) -> Union[Tuple[Optional[float], Optional[float]], PhotometryBox]:
        """
        Function for calculating the average flux density for the
        ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Model parameters and values.
        synphot : SyntheticPhotometry, None
            Synthetic photometry object. The object is created if set
            to ``None``.
        return_box : bool
            Return a :class:`~species.core.box.PhotometryBox`
            if set to ``True`` or return the two values that are
            specified below if set to ``False``. By default, the
            argument is set to ``False``. The advantage of
            returning the output in a
            :class:`~species.core.box.PhotometryBox` is that it can
            directly be provided as input to
            :func:`~species.plot.plot_spectrum.plot_spectrum`.

        Returns
        -------
        float
            Average flux (W m-2 um-1).
        float, None
            Uncertainty (W m-2 um-1), which is set to ``None``.
        """

        param_list = self.get_parameters()

        for key in param_list:
            if key not in model_param.keys():
                raise ValueError(
                    f"The '{key}' parameter is required by '{self.model}'. "
                    f"The mandatory parameters are {param_list}."
                )

        if self.spectrum_interp is None:
            self.interpolate_grid()

        model_box = self.get_model(model_param)

        if synphot is None:
            synphot = SyntheticPhotometry(self.filter_name)

        model_flux = synphot.spectrum_to_flux(model_box.wavelength, model_box.flux)

        if return_box:
            model_mag = self.get_magnitude(model_param)

            phot_box = create_box(
                boxtype="photometry",
                name=self.model,
                wavelength=[self.mean_wavelength],
                flux=[model_flux],
                app_mag=[(model_mag[0], None)],
                abs_mag=[(model_mag[1], None)],
                filter_name=[self.filter_name],
            )

            return phot_box

        return model_flux

    @typechecked
    def get_magnitude(
        self,
        model_param: Dict[str, float],
        return_box: bool = False,
    ) -> Union[Tuple[Optional[float], Optional[float]], PhotometryBox]:
        """
        Function for calculating the apparent and absolute magnitudes
        for the ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters. A ``radius`` (Rjup),
            and ``parallax`` (mas) or ``distance`` (pc) are required
            for the apparent magnitude (i.e. to scale the flux from
            the planet to the observer). Only a ``radius`` is
            required for the absolute magnitude.
        return_box : bool
            Return a :class:`~species.core.box.PhotometryBox`
            if set to ``True`` or return the two values that are
            specified below if set to ``False``. By default, the
            argument is set to ``False``. The advantage of
            returning the output in a
            :class:`~species.core.box.PhotometryBox` is that it can
            directly be provided as input to
            :func:`~species.plot.plot_spectrum.plot_spectrum`.

        Returns
        -------
        float
            Apparent magnitude. A ``None`` is returned if the
            dictionary of ``model_param`` does not contain a
            ``radius``, and ``parallax`` or ``distance``.
        float, None
            Absolute magnitude. A ``None`` is returned if the
            dictionary of ``model_param`` does not contain a
            ``radius``.
        """

        param_list = self.get_parameters()

        for key in param_list:
            if key not in model_param.keys():
                raise ValueError(
                    f"The '{key}' parameter is required by '{self.model}'. "
                    f"The mandatory parameters are {param_list}."
                )

        if self.spectrum_interp is None:
            self.interpolate_grid()

        try:
            spectrum = self.get_model(model_param)

        except ValueError:
            warnings.warn(
                f"The set of model parameters {model_param} is outside "
                f"the grid range {self.get_bounds()} so returning a NaN."
            )

            return np.nan, np.nan

        if spectrum.wavelength.size == 0:
            app_mag = (np.nan, None)
            abs_mag = (np.nan, None)

        else:
            synphot = SyntheticPhotometry(self.filter_name)

            if "radius" in model_param and "parallax" in model_param:
                app_mag, abs_mag = synphot.spectrum_to_magnitude(
                    spectrum.wavelength,
                    spectrum.flux,
                    parallax=(model_param["parallax"], None),
                )

            elif "radius" in model_param and "distance" in model_param:
                app_mag, abs_mag = synphot.spectrum_to_magnitude(
                    spectrum.wavelength,
                    spectrum.flux,
                    distance=(model_param["distance"], None),
                )

            else:
                app_mag = (None, None)
                abs_mag = (None, None)

                if "radius" in model_param:
                    distance = 10.0  # (pc)

                    spectrum.flux *= (model_param["radius"] * constants.R_JUP) ** 2
                    spectrum.flux /= (distance * constants.PARSEC) ** 2

                    _, abs_mag = synphot.spectrum_to_magnitude(
                        spectrum.wavelength, spectrum.flux, distance=(distance, None)
                    )

        if return_box:
            model_flux = self.get_flux(model_param)

            phot_box = create_box(
                boxtype="photometry",
                name=self.model,
                wavelength=[self.mean_wavelength],
                flux=[model_flux],
                app_mag=[(app_mag[0], None)],
                abs_mag=[(abs_mag[0], None)],
                filter_name=[self.filter_name],
            )

            return phot_box

        return app_mag[0], abs_mag[0]

    @typechecked
    def get_bounds(self) -> Dict[str, Tuple[float, float]]:
        """
        Function for extracting the grid boundaries.

        Returns
        -------
        dict
            Boundaries of parameter grid.
        """

        param_list = self.get_parameters()

        with self.open_database() as hdf5_file:
            bounds = {}

            for param_item in param_list:
                data = hdf5_file[f"models/{self.model}/{param_item}"]
                bounds[param_item] = (data[0], data[-1])

        return bounds

    @typechecked
    def get_wavelengths(self) -> np.ndarray:
        """
        Function for extracting the wavelength points.

        Returns
        -------
        np.ndarray
            Wavelength points (um).
        """

        with self.open_database() as hdf5_file:
            wavelength = np.array(hdf5_file[f"models/{self.model}/wavelength"])

        return wavelength

    @typechecked
    def get_points(self) -> Dict[str, np.ndarray]:
        """
        Function for extracting the grid points.

        Returns
        -------
        dict
            Parameter points of the model grid.
        """

        param_list = self.get_parameters()

        points = {}

        with self.open_database() as hdf5_file:
            points = {}

            for param_item in param_list:
                data = hdf5_file[f"models/{self.model}/{param_item}"]
                points[param_item] = np.array(data)

        return points

    @typechecked
    def get_parameters(self) -> List[str]:
        """
        Function for extracting the parameter names.

        Returns
        -------
        list(str)
            Model parameters.
        """

        with self.open_database() as hdf5_file:
            dset = hdf5_file[f"models/{self.model}"]

            if "n_param" in dset.attrs:
                n_param = dset.attrs["n_param"]
            else:
                n_param = dset.attrs["nparam"]

            param = []
            for i in range(n_param):
                param.append(dset.attrs[f"parameter{i}"])

        return param

    @typechecked
    def get_sampling(self) -> float:
        """
        Function for returning the wavelength sampling,
        :math:`\\lambda/\\Delta\\lambda`, of the
        model spectra as stored in the database.

        Returns
        -------
        float
            Wavelength sampling, :math:`\\lambda/\\Delta\\lambda`.
        """

        wavel_points = self.get_wavelengths()

        wavel_mean = (wavel_points[1:] + wavel_points[:-1]) / 2.0
        wavel_sampling = wavel_mean / np.diff(wavel_points)

        return np.mean(wavel_sampling)

    @typechecked
    def binary_spectrum(
        self,
        model_param: Dict[str, float],
        spec_res: Optional[float] = None,
        wavel_resample: Optional[np.ndarray] = None,
        **kwargs,
    ) -> ModelBox:
        """
        Function for extracting a model spectrum of a binary system.
        A weighted combination of two spectra will be returned. The
        ``model_param`` dictionary should contain the parameters
        for both components (e.g. ``teff_0`` and ``teff_1``, instead
        of ``teff``). Apart from that, the same parameters are used
        as with :func:`~species.read.read_model.ReadModel.get_model`.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The grid
            boundaries of the spectra in the database can be obtained
            with
            :func:`~species.read.read_model.ReadModel.get_bounds()`.
        spec_res : float, None
            Spectral resolution that is used for smoothing the
            spectrum with a Gaussian kernel. No smoothing is applied
            if the argument is set to ``None``.
        wavel_resample : np.ndarray, None
            Wavelength points (um) to which the spectrum is resampled.
            In that case, ``spec_res`` can still be used for smoothing
            the spectrum with a Gaussian kernel. The original
            wavelength points are used if the argument is set to
            ``None``.

        Returns
        -------
        species.core.box.ModelBox
            Box with the model spectrum.
        """

        if "smooth" in kwargs:
            warnings.warn(
                "The 'smooth' parameter has been "
                "deprecated. Please set only the "
                "'spec_res' argument, which can be set "
                "to None for not applying a smoothing.",
                DeprecationWarning,
            )

            if not kwargs["smooth"] and spec_res is not None:
                spec_res = None

        # Get grid boundaries

        param_0 = binary_to_single(model_param, 0)

        model_box_0 = self.get_model(
            param_0,
            spec_res=spec_res,
            wavel_resample=wavel_resample,
        )

        param_1 = binary_to_single(model_param, 1)

        model_box_1 = self.get_model(
            param_1,
            spec_res=spec_res,
            wavel_resample=wavel_resample,
        )

        # Weighted flux of two spectra for atmospheric asymmetries
        # Or simply the same in case of an actual binary system

        if "spec_weight" in model_param:
            flux_comb = (
                model_param["spec_weight"] * model_box_0.flux
                + (1.0 - model_param["spec_weight"]) * model_box_1.flux
            )

        else:
            flux_comb = model_box_0.flux + model_box_1.flux

        model_box = create_box(
            boxtype="model",
            model=self.model,
            wavelength=model_box_0.wavelength,
            flux=flux_comb,
            parameters=model_param,
            quantity="flux",
            spec_res=spec_res,
        )

        return model_box

    @typechecked
    def integrate_spectrum(
        self, model_param: Dict[str, float]
    ) -> Tuple[float, Optional[float]]:
        """
        Function for calculating the bolometric flux by integrating
        a model spectrum at the requested parameters. Therefore, when
        extinction is applied to the spectrum, the luminosity is the
        extinct luminosity and not the intrinsic luminosity. Without
        applying extinction, the integrated luminosity should in
        principle be the same as the luminosity calculated directly
        from the :math:`T_\\mathrm{eff}` and radius parameters, unless
        the radiative-convective model had not fully converged for a
        particular set of input parameters. It can thus be useful
        to check if the integrated luminosity is indeed consistent
        with the :math:`T_\\mathrm{eff}` of the model.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The grid
            boundaries of the spectra in the database can be obtained
            with
            :func:`~species.read.read_model.ReadModel.get_bounds()`.

        Returns
        -------
        float
            Effective temperature (K) calculated with the
            Stefan–Boltzmann law from the integrated flux.
        float, None
            Bolometric luminosity (:math:`\\log{(L/L_\\odot)}`).
            The returned value is set to ``None`` in case
            the ``model_param`` dictionary does not contain
            the radius parameter.
        """

        wavel_points = self.get_wavelengths()

        if self.wavel_range is None:
            self.wavel_range = (wavel_points[0], wavel_points[-1])

        if (
            self.wavel_range[0] != wavel_points[0]
            or self.wavel_range[1] != wavel_points[-1]
        ):
            warnings.warn(
                "The 'wavel_range' is not set to the maximum "
                "available range. To maximize the accuracy when "
                "calculating the bolometric luminosity, it is "
                "recommended to set 'wavel_range=None'."
            )

        param_copy = model_param.copy()

        if "parallax" in param_copy:
            del param_copy["parallax"]

        if "distance" in param_copy:
            del param_copy["distance"]

        model_box = self.get_model(param_copy)

        flux_int = simpson(y=model_box.flux, x=model_box.wavelength)

        if "radius" in param_copy:
            bol_lum = (
                4.0 * np.pi * (param_copy["radius"] * constants.R_JUP) ** 2 * flux_int
            )
            log_lum = np.log10(bol_lum / constants.L_SUN)
        else:
            log_lum = None

            warnings.warn(
                "Please include the 'radius' parameter in the "
                "'model_param' dictionary, if the bolometric "
                "luminosity should be calculated."
            )

        teff_int = (flux_int / constants.SIGMA_SB) ** 0.25

        return teff_int, log_lum

    @typechecked
    def create_color_magnitude(
        self,
        model_param: Dict[str, float],
        filters_color: Tuple[str, str],
        filter_mag: str,
    ) -> ColorMagBox:
        """
        Function for creating a :class:`~species.core.box.
        ColorMagBox` for a given set of filter names and model
        parameters. The effective temperature, :math:`T_\\mathrm{eff}`,
        is varied such that the returned :class:`~species.core.box.
        ColorMagBox` contains the colors as function of
        :math:`T_\\mathrm{eff}` and can be provide as input to the
        :func:`~species.plot.plot_color.plot_color_magnitude` function.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The boundaries
            of the model grid can be inspected by using the
            :func:`~species.read.read_model.ReadModel.get_bounds()`
            method. The effective temperature, :math:`T_\\mathrm{eff}`,
            does not need to be included in the dictionary since it
            is varied. The values of :math:`T_\\mathrm{eff}` are set to
            the grid points. The grid points can be inspected with the
            :func:`~species.read.read_model.ReadModel.get_points()`
            method.
        filters_color : tuple(str, str)
            Filter names that are used for the color. Any of
            the filter names from the `SVO Filter Profile Service
            <http://svo2.cab.inta-csic.es/svo/theory/fps/>`_ are
            compatible.
        filter_mag : str
            Filter name that is used for the magnitude. Any of
            the filter names from the `SVO Filter Profile Service
            <http://svo2.cab.inta-csic.es/svo/theory/fps/>`_ are
            compatible.

        Returns
        -------
        species.core.box.ColorMagBox
            Box with the colors and magnitudes.
        """

        if "distance" not in model_param:
            model_param["distance"] = 10.0

        if "radius" not in model_param:
            model_param["radius"] = 1.0

        if "parallax" in model_param:
            del model_param["parallax"]

        read_model_1 = ReadModel(self.model, filter_name=filters_color[0])
        read_model_2 = ReadModel(self.model, filter_name=filters_color[0])
        read_model_3 = ReadModel(self.model, filter_name=filter_mag)

        model_points = self.get_points()

        param_list = []
        color_list = []
        mag_list = []

        for param_item in model_points["teff"]:
            model_param["teff"] = param_item

            mag_1 = read_model_1.get_magnitude(model_param)
            mag_2 = read_model_2.get_magnitude(model_param)
            mag_3 = read_model_3.get_magnitude(model_param)

            param_list.append(param_item)
            color_list.append(mag_1[0] - mag_2[0])
            mag_list.append(mag_3[0])

        return create_box(
            "colormag",
            library=self.model,
            object_type="spectra",
            filters_color=filters_color,
            filter_mag=filter_mag,
            color=color_list,
            magnitude=mag_list,
            sptype=param_list,
        )

    @typechecked
    def create_color_color(
        self,
        model_param: Dict[str, float],
        filters_colors: Tuple[Tuple[str, str], Tuple[str, str]],
    ) -> ColorColorBox:
        """
        Function for creating a :class:`~species.core.box.
        ColorColorBox` for a given set of filter names and model
        parameters. The effective temperature, :math:`T_\\mathrm{eff}`,
        is varied such that the returned :class:`~species.core.box.
        ColorColorBox` contains the colors as function of
        :math:`T_\\mathrm{eff}` and can be provide as input to the
        :func:`~species.plot.plot_color.plot_color_color` function.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The boundaries
            of the model grid can be inspected by using the
            :func:`~species.read.read_model.ReadModel.get_bounds()`
            method. The effective temperature, :math:`T_\\mathrm{eff}`,
            does not need to be included in the dictionary since it
            is varied. The values of :math:`T_\\mathrm{eff}` are set to
            the grid points. The grid points can be inspected with the
            :func:`~species.read.read_model.ReadModel.get_points()`
            method.
        filters_colors : tuple(tuple(str, str), tuple(str, str))
            Filter names that are used for the two colors. Any of
            the filter names from the `SVO Filter Profile Service
            <http://svo2.cab.inta-csic.es/svo/theory/fps/>`_ are
            compatible.

        Returns
        -------
        species.core.box.ColorColorBox
            Box with the colors.
        """

        if "distance" not in model_param:
            model_param["distance"] = 10.0

        if "radius" not in model_param:
            model_param["radius"] = 1.0

        if "parallax" in model_param:
            del model_param["parallax"]

        read_model_1 = ReadModel(self.model, filter_name=filters_colors[0][0])
        read_model_2 = ReadModel(self.model, filter_name=filters_colors[0][1])
        read_model_3 = ReadModel(self.model, filter_name=filters_colors[1][0])
        read_model_4 = ReadModel(self.model, filter_name=filters_colors[1][1])

        model_points = self.get_points()

        param_list = []
        color_1_list = []
        color_2_list = []

        for param_item in model_points["teff"]:
            model_param["teff"] = param_item

            mag_1 = read_model_1.get_magnitude(model_param)
            mag_2 = read_model_2.get_magnitude(model_param)
            mag_3 = read_model_3.get_magnitude(model_param)
            mag_4 = read_model_4.get_magnitude(model_param)

            param_list.append(param_item)
            color_1_list.append(mag_1[0] - mag_2[0])
            color_2_list.append(mag_3[0] - mag_4[0])

        return create_box(
            "colorcolor",
            library=self.model,
            object_type="spectra",
            filters=filters_colors,
            color1=color_1_list,
            color2=color_2_list,
            sptype=param_list,
        )

    # def test_accuracy(self):
    #     with self.open_database() as hdf5_file:
    #         wl_points = np.array(hdf5_file[f"models/{self.model}/wavelength"])
    #         grid_flux = np.array(hdf5_file[f"models/{self.model}/flux"])
    #
    #     import matplotlib.pyplot as plt
    #
    #     for i in range(grid_flux.shape[0]):
    #         if i == 0:
    #             continue
    #         if i == grid_flux.shape[0]-1:
    #             continue
    #
    #         # Get the grid points
    #
    #         grid_points = self.get_points()
    #
    #         # Select the required Teff points of the grid
    #
    #         # if "teff" in grid_points and teff_range is not None:
    #         #     teff_select = (teff_range[0] <= grid_points["teff"]) & (
    #         #         grid_points["teff"] <= teff_range[1]
    #         #     )
    #         #
    #         #     grid_points["teff"] = grid_points["teff"][teff_select]
    #         #
    #         # else:
    #         #     teff_select = np.ones(grid_points["teff"].size, dtype=bool)
    #
    #         # Create list with grid points
    #
    #         grid_points = list(grid_points.values())
    #
    #         # Get the boolean array for selecting the fluxes
    #         # within the requested wavelength range
    #
    #         # if self.wl_index is None:
    #         #     self.wl_points, self.wl_index = self.wavelength_points()
    #
    #         # Open de HDF5 database and read the model fluxes
    #
    #         # with self.open_database() as hdf5_file:
    #         #     grid_flux = np.array(hdf5_file[f"models/{self.model}/flux"])
    #         #     grid_flux = grid_flux[..., self.wl_index]
    #         #     grid_flux = grid_flux[teff_select, ...]
    #
    #         grid_points_select = grid_points.copy()
    #         grid_points_select[0] = np.delete(grid_points_select[0], i)
    #
    #         grid_flux_select = np.delete(grid_flux, i, axis=0)
    #
    #         # Interpolate the grid of model spectra
    #
    #         spectrum_interp = RegularGridInterpolator(
    #             grid_points_select,
    #             grid_flux_select,
    #             method="linear",
    #             bounds_error=False,
    #             fill_value=np.nan,
    #         )
    #
    #         param_test = (grid_points[0][i], grid_points[1][0])
    #         flux_interp = spectrum_interp(param_test)
    #         flux_true = grid_flux[i, 0]
    #         diff = (flux_true-flux_interp)/flux_true
    #         # plt.figure(figsize=(10., 3.))
    #         # plt.plot(wl_points, 100.*diff, '-', lw=0.3)
    #         # plt.yscale('log')
    #         # plt.show()
    #         # exit()
    #
    #         # grid_flux = grid_flux[..., self.wl_index]
    #         # grid_flux = grid_flux[teff_select, ...]



"""
Utility functions for data processing.
"""

import os
import tarfile
import warnings

from pathlib import Path
from typing import List, Optional, Tuple

import h5py
import numpy as np

from scipy.interpolate import griddata
from typeguard import typechecked

from species.core import constants


@typechecked
def extract_tarfile(
    data_file: str,
    data_folder: str,
    member_list: Optional[List[tarfile.TarInfo]] = None,
) -> None:
    """
    Function for safely unpacking a TAR file (`see details
    <https://github.com/advisories/GHSA-gw9q-c7gh-j9vm>`_.

    Parameters
    ----------
    data_file : str
        Path of the TAR file.
    data_folder : str
        Path of the data folder where the TAR file will be extracted.
    member_list : list(tarfile.TarInfo), None
        List with filenames that should be extracted from the TAR file.
        All files are extracted if the argument is set to ``None``.

    Returns
    -------
    NoneType
        None
    """

    with tarfile.open(data_file) as tar:
        if member_list is None:
            member_list = tar.getmembers()

        @typechecked
        def is_within_directory(directory: str, target: str) -> bool:
            abs_directory = os.path.abspath(directory)
            abs_target = os.path.abspath(target)

            prefix = os.path.commonprefix([abs_directory, abs_target])

            return prefix == abs_directory

        @typechecked
        def safe_extract(
            tar: tarfile.TarFile,
            path: str = ".",
            member_list: Optional[List] = None,
            numeric_owner: bool = False,
        ) -> None:
            for member in member_list:
                member_path = Path(path) / Path(member.name)

                if not is_within_directory(path, str(member_path)):
                    raise Exception("Attempted path traversal in TAR file")

            tar.extractall(
                path, members=member_list, numeric_owner=numeric_owner, filter="data"
            )

        safe_extract(tar, data_folder, member_list=member_list)


@typechecked
def update_sptype(sptypes: np.ndarray) -> List[str]:
    """
    Function to update a list with spectral types to two characters
    (e.g., M8, L3, or T1). The spectral to is set to NaN in case the
    first character is not recognized or the second character is not
    a numerical value.

    Parameters
    ----------
    sptypes : np.ndarray
        Input spectral types.

    Returns
    -------
    list(str)
        Output spectral types.
    """

    sptype_list = ["O", "B", "A", "F", "G", "K", "M", "L", "T", "Y"]

    sptypes_updated = []

    for spt_item in sptypes:
        if spt_item == "None":
            sptypes_updated.append("None")

        elif spt_item == "null":
            sptypes_updated.append("None")

        else:
            if (
                len(spt_item) > 1
                and spt_item[0] in sptype_list
                and spt_item[1].isnumeric()
            ):
                sptypes_updated.append(spt_item[:2])

            else:
                sptypes_updated.append("None")

    return sptypes_updated


def update_filter(filter_in):
    """
    Function to update a filter ID from the Vizier Photometry viewer
    VOTable to the filter ID from the SVO Filter Profile Service.

    Parameters
    ----------
    filter_in : str
        Filter ID in the format of the Vizier Photometry viewer.

    Returns
    -------
    str
        Filter ID in the format of the SVO Filter Profile Service.
    """

    if filter_in[0:5] == b"2MASS":
        filter_out = str(b"2MASS/2MASS." + filter_in[6:])

    elif filter_in[0:4] == b"WISE":
        filter_out = str(b"WISE/WISE." + filter_in[5:])

    elif filter_in[0:10] == b"GAIA/GAIA2":
        filter_out = str(filter_in[0:9] + b"0" + filter_in[10:])

    else:
        filter_out = None

    return filter_out


@typechecked
def sort_data(
    param_teff: np.ndarray,
    param_logg: Optional[np.ndarray],
    param_feh: Optional[np.ndarray],
    param_co: Optional[np.ndarray],
    param_fsed: Optional[np.ndarray],
    param_logkzz: Optional[np.ndarray],
    param_adindex: Optional[np.ndarray],
    param_logcoiso: Optional[np.ndarray],
    wavelength: np.ndarray,
    flux: np.ndarray,
) -> List[np.ndarray]:
    """
    Parameters
    ----------
    param_teff : np.ndarray
        Array with the effective temperature (K) of each spectrum.
    param_logg : np.ndarray, None
        Array with the log10 of the surface gravity (cm s-2) of each
        spectrum.
    param_feh : np.ndarray, None
        Array with the metallicity of each spectrum. Not used if set
        to ``None``.
    param_co : np.ndarray, None
        Array with the carbon-to-oxygen ratio of each spectrum. Not
        used if set to ``None``.
    param_fsed : np.ndarray, None
        Array with the sedimentation parameter of each spectrum. Not
        used if set to ``None``.
    param_logkzz : np.ndarray, None
        Array with the log10 of the eddy diffusion coefficient
        (cm2 s-1) of each spectrum. Not used if set to ``None``.
    param_adindex : np.ndarray, None
        Array with the effective adiabatic index. Not used if
        set to ``None``.
    param_logcoiso : np.ndarray, None
        Array with the 12CO/13CO isotopologue ratio. Not used if
        set to ``None``.
    wavelength : np.ndarray
        Array with the wavelengths (um).
    flux : np.ndarray
        Array with the spectra with dimensions
        ``(n_spectra, n_wavelengths)``.

    Returns
    -------
    list(np.ndarray, )
        List with the unique values of the atmosphere parameters (each
    in a separate array), an
        array with the wavelengths, and a multidimensional array with
        the sorted spectra.
    """

    n_spectra = param_teff.shape[0]

    teff_unique = np.unique(param_teff)
    spec_shape = [teff_unique.shape[0]]

    print("\nGrid points stored in the database:")
    print(f"   - Teff = {teff_unique}")

    if param_logg is not None:
        logg_unique = np.unique(param_logg)
        spec_shape.append(logg_unique.shape[0])
        print(f"   - log(g) = {logg_unique}")

    if param_feh is not None:
        feh_unique = np.unique(param_feh)
        spec_shape.append(feh_unique.shape[0])
        print(f"   - [Fe/H] = {feh_unique}")

    if param_co is not None:
        co_unique = np.unique(param_co)
        spec_shape.append(co_unique.shape[0])
        print(f"   - C/O = {co_unique}")

    if param_fsed is not None:
        fsed_unique = np.unique(param_fsed)
        spec_shape.append(fsed_unique.shape[0])
        print(f"   - f_sed = {fsed_unique}")

    if param_logkzz is not None:
        logkzz_unique = np.unique(param_logkzz)
        spec_shape.append(logkzz_unique.shape[0])
        print(f"   - log(Kzz) = {logkzz_unique}")

    if param_adindex is not None:
        adindex_unique = np.unique(param_adindex)
        spec_shape.append(adindex_unique.shape[0])
        print(f"   - gamma_ad = {adindex_unique}")

    if param_logcoiso is not None:
        logcoiso_unique = np.unique(param_logcoiso)
        spec_shape.append(logcoiso_unique.shape[0])
        print(f"   - log(12CO/13CO) = {logcoiso_unique}")

    spec_shape.append(wavelength.shape[0])

    spectrum = np.zeros(spec_shape)

    for i in range(n_spectra):
        # The parameter order is: Teff, log(g), [Fe/H], C/O,
        # f_sed, log(Kzz), ad_index. Not all parameters have
        # to be included but the order matters

        index_teff = np.argwhere(teff_unique == param_teff[i])[0][0]
        spec_select = [index_teff]

        if param_logg is not None:
            index_logg = np.argwhere(logg_unique == param_logg[i])[0][0]
            spec_select.append(index_logg)

        if param_feh is not None:
            index_feh = np.argwhere(feh_unique == param_feh[i])[0][0]
            spec_select.append(index_feh)

        if param_co is not None:
            index_co = np.argwhere(co_unique == param_co[i])[0][0]
            spec_select.append(index_co)

        if param_fsed is not None:
            index_fsed = np.argwhere(fsed_unique == param_fsed[i])[0][0]
            spec_select.append(index_fsed)

        if param_logkzz is not None:
            index_logkzz = np.argwhere(logkzz_unique == param_logkzz[i])[0][0]
            spec_select.append(index_logkzz)

        if param_adindex is not None:
            index_adindex = np.argwhere(adindex_unique == param_adindex[i])[0][0]
            spec_select.append(index_adindex)

        if param_logcoiso is not None:
            index_logcoiso = np.argwhere(logcoiso_unique == param_logcoiso[i])[0][0]
            spec_select.append(index_logcoiso)

        spec_select.append(...)

        spectrum[tuple(spec_select)] = flux[i]

    sorted_data = [teff_unique]

    if param_logg is not None:
        sorted_data.append(logg_unique)

    if param_feh is not None:
        sorted_data.append(feh_unique)

    if param_co is not None:
        sorted_data.append(co_unique)

    if param_fsed is not None:
        sorted_data.append(fsed_unique)

    if param_logkzz is not None:
        sorted_data.append(logkzz_unique)

    if param_adindex is not None:
        sorted_data.append(adindex_unique)

    if param_logcoiso is not None:
        sorted_data.append(logcoiso_unique)

    sorted_data.append(wavelength)
    sorted_data.append(spectrum)

    return sorted_data


@typechecked
def write_data(
    model: str,
    parameters: List[str],
    wavel_sampling: float,
    database: h5py._hl.files.File,
    data_sorted: List[np.ndarray],
) -> None:
    """
    Function for writing the model spectra and parameters to the
    database.

    Parameters
    ----------
    model : str
        Atmosphere model.
    parameters : list(str)
        Model parameters.
    wavel_sampling : float
        Wavelength sampling :math:`\\lambda/\\Delta\\lambda`.
    database: h5py._hl.files.File
        HDF5 database.
    data_sorted : list(np.ndarray)
        Sorted model data with the parameter values, wavelength
        points (um), and flux densities (W m-2 um-1).

    Returns
    -------
    NoneType
        None
    """

    n_param = len(parameters)

    if f"models/{model}" in database:
        del database[f"models/{model}"]

    dset = database.create_group(f"models/{model}")

    dset.attrs["n_param"] = n_param
    dset.attrs["lambda/d_lambda"] = wavel_sampling

    for i, item in enumerate(parameters):
        dset.attrs[f"parameter{i}"] = item

        database.create_dataset(f"models/{model}/{item}", data=data_sorted[i])

    database.create_dataset(f"models/{model}/wavelength", data=data_sorted[n_param])

    database.create_dataset(f"models/{model}/flux", data=data_sorted[n_param + 1])


@typechecked
def add_missing(
    model: str, parameters: List[str], database: h5py._hl.files.File
) -> None:
    """
    Function for adding missing grid points with a linear
    interpolation.

    Parameters
    ----------
    model : str
        Atmosphere model.
    parameters : list(str)
        Model parameters.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    NoneType
        None
    """

    print("\nNumber of grid points per parameter:")

    grid_shape = []
    param_data = []

    for i, item in enumerate(parameters):
        grid_shape.append(database[f"models/{model}/{item}"].shape[0])
        param_data.append(np.asarray(database[f"models/{model}/{item}"]))
        print(f"   - {item}: {grid_shape[i]}")

    flux = np.asarray(database[f"models/{model}/flux"])  # (W m-1 um-1)
    # Should be < and not <= since missing spectra have zero fluxes
    flux[flux < 0.0] = 1e-50
    flux = np.log10(flux)

    count_total = 0
    count_interp = 0
    count_missing = 0

    if not np.isfinite(np.sum(flux)):
        print("\nFix missing grid points with a linear interpolation:")

    if len(parameters) == 1:
        # Blackbody spectra
        pass

    elif len(parameters) == 2:
        find_missing = np.zeros(grid_shape, dtype=bool)

        values = []
        points = [[], []]
        new_points = [[], []]

        for i in range(grid_shape[0]):
            for j in range(grid_shape[1]):
                if np.isinf(np.sum(flux[i, j, ...])):
                    print("   - ", end="")
                    print(f"{parameters[0]} = {param_data[0][i]}, ", end="")
                    print(f"{parameters[1]} = {param_data[1][j]}")

                    if 0 < i < grid_shape[0] - 1:
                        check_low = np.isinf(np.sum(flux[i - 1, j, ...]))
                        check_up = np.isinf(np.sum(flux[i + 1, j, ...]))

                        # Linear scaling of the intermediate Teff point
                        scaling = (param_data[0][i] - param_data[0][i - 1]) / (
                            param_data[0][i + 1] - param_data[0][i - 1]
                        )

                        if not check_low and not check_up:
                            flux_low = flux[i - 1, j, ...]
                            flux_up = flux[i + 1, j, ...]
                            flux[i, j, ...] = (
                                flux_low * (1.0 - scaling) + flux_up * scaling
                            )
                            count_interp += 1

                        else:
                            find_missing[i, j] = True

                    else:
                        find_missing[i, j] = True

                else:
                    points[0].append(param_data[0][i])
                    points[1].append(param_data[1][j])

                    values.append(flux[i, j, ...])

                new_points[0].append(param_data[0][i])
                new_points[1].append(param_data[1][j])

                count_total += 1

        values = np.asarray(values)
        points = np.asarray(points)
        new_points = np.asarray(new_points)

        if np.sum(find_missing) > 0:
            flux_int = griddata(
                points.T, values, new_points.T, method="linear", fill_value=np.nan
            )

            count = 0

            for i in range(grid_shape[0]):
                for j in range(grid_shape[1]):
                    if np.isnan(np.sum(flux_int[count, :])):
                        count_missing += 1

                    elif np.isinf(np.sum(flux[i, j, ...])):
                        flux[i, j, :] = flux_int[count, :]
                        count_interp += 1

                    count += 1

            if count_missing > 0:
                warnings.warn(
                    f"Could not interpolate {count_missing} grid "
                    "points so storing zeros instead."
                )

                print("\nThe following grid points are missing:")

                for i in range(flux_int.shape[0]):
                    if np.isnan(np.sum(flux_int[i, :])):
                        print("   - ", end="")
                        print(f"{parameters[0]} = {new_points[0][i]}, ", end="")
                        print(f"{parameters[1]} = {new_points[1][i]}")

    elif len(parameters) == 3:
        find_missing = np.zeros(grid_shape, dtype=bool)

        values = []
        points = [[], [], []]
        new_points = [[], [], []]

        for i in range(grid_shape[0]):
            for j in range(grid_shape[1]):
                for k in range(grid_shape[2]):
                    if np.isinf(np.sum(flux[i, j, k, ...])):
                        print("   - ", end="")
                        print(f"{parameters[0]} = {param_data[0][i]}, ", end="")
                        print(f"{parameters[1]} = {param_data[1][j]}, ", end="")
                        print(f"{parameters[2]} = {param_data[2][k]}")

                        if 0 < i < grid_shape[0] - 1:
                            check_low = np.isinf(np.sum(flux[i - 1, j, k, ...]))
                            check_up = np.isinf(np.sum(flux[i + 1, j, k, ...]))

                            # Linear scaling of the intermediate Teff point
                            scaling = (param_data[0][i] - param_data[0][i - 1]) / (
                                param_data[0][i + 1] - param_data[0][i - 1]
                            )

                            if not check_low and not check_up:
                                flux_low = flux[i - 1, j, k, ...]
                                flux_up = flux[i + 1, j, k, ...]
                                flux[i, j, k, ...] = (
                                    flux_low * (1.0 - scaling) + flux_up * scaling
                                )
                                count_interp += 1

                            else:
                                find_missing[i, j, k] = True

                        else:
                            find_missing[i, j, k] = True

                    else:
                        points[0].append(param_data[0][i])
                        points[1].append(param_data[1][j])
                        points[2].append(param_data[2][k])

                        values.append(flux[i, j, k, ...])

                    new_points[0].append(param_data[0][i])
                    new_points[1].append(param_data[1][j])
                    new_points[2].append(param_data[2][k])

                    count_total += 1

        values = np.asarray(values)
        points = np.asarray(points)
        new_points = np.asarray(new_points)

        if np.sum(find_missing) > 0:
            flux_int = griddata(
                points.T, values, new_points.T, method="linear", fill_value=np.nan
            )

            count = 0

            for i in range(grid_shape[0]):
                for j in range(grid_shape[1]):
                    for k in range(grid_shape[2]):
                        if np.isnan(np.sum(flux_int[count, :])):
                            count_missing += 1

                        elif np.isinf(np.sum(flux[i, j, k, ...])):
                            flux[i, j, k, :] = flux_int[count, :]
                            count_interp += 1

                        count += 1

            if count_missing > 0:
                warnings.warn(
                    f"Could not interpolate {count_missing} grid "
                    "points so storing zeros instead."
                )

                print("\nThe following grid points are missing:")

                for i in range(flux_int.shape[0]):
                    if np.isnan(np.sum(flux_int[i, :])):
                        print("   - ", end="")
                        print(f"{parameters[0]} = {new_points[0][i]}, ", end="")
                        print(f"{parameters[1]} = {new_points[1][i]}, ", end="")
                        print(f"{parameters[2]} = {new_points[2][i]}")

    elif len(parameters) == 4:
        find_missing = np.zeros(grid_shape, dtype=bool)

        values = []
        points = [[], [], [], []]
        new_points = [[], [], [], []]

        for i in range(grid_shape[0]):
            for j in range(grid_shape[1]):
                for k in range(grid_shape[2]):
                    for m in range(grid_shape[3]):
                        if np.isinf(np.sum(flux[i, j, k, m, ...])):
                            print("   - ", end="")
                            print(f"{parameters[0]} = {param_data[0][i]}, ", end="")
                            print(f"{parameters[1]} = {param_data[1][j]}, ", end="")
                            print(f"{parameters[2]} = {param_data[2][k]}, ", end="")
                            print(f"{parameters[3]} = {param_data[3][m]}")

                            if 0 < i < grid_shape[0] - 1:
                                check_low = np.isinf(np.sum(flux[i - 1, j, k, m, ...]))
                                check_up = np.isinf(np.sum(flux[i + 1, j, k, m, ...]))

                                # Linear scaling of the intermediate Teff point
                                scaling = (param_data[0][i] - param_data[0][i - 1]) / (
                                    param_data[0][i + 1] - param_data[0][i - 1]
                                )

                                if not check_low and not check_up:
                                    flux_low = flux[i - 1, j, k, m, ...]
                                    flux_up = flux[i + 1, j, k, m, ...]
                                    flux[i, j, k, m, ...] = (
                                        flux_low * (1.0 - scaling) + flux_up * scaling
                                    )
                                    count_interp += 1

                                else:
                                    find_missing[i, j, k, m] = True

                            else:
                                find_missing[i, j, k, m] = True

                        else:
                            points[0].append(param_data[0][i])
                            points[1].append(param_data[1][j])
                            points[2].append(param_data[2][k])
                            points[3].append(param_data[3][m])

                            values.append(flux[i, j, k, m, ...])

                        new_points[0].append(param_data[0][i])
                        new_points[1].append(param_data[1][j])
                        new_points[2].append(param_data[2][k])
                        new_points[3].append(param_data[3][m])

                        count_total += 1

        values = np.asarray(values)
        points = np.asarray(points)
        new_points = np.asarray(new_points)

        if np.sum(find_missing) > 0:
            flux_int = griddata(
                points.T, values, new_points.T, method="linear", fill_value=np.nan
            )

            count = 0

            for i in range(grid_shape[0]):
                for j in range(grid_shape[1]):
                    for k in range(grid_shape[2]):
                        for m in range(grid_shape[3]):
                            if np.isnan(np.sum(flux_int[count, :])):
                                count_missing += 1

                            elif np.isinf(np.sum(flux[i, j, k, m, ...])):
                                flux[i, j, k, m, :] = flux_int[count, :]
                                count_interp += 1

                            count += 1

            if count_missing > 0:
                warnings.warn(
                    f"Could not interpolate {count_missing} grid "
                    "points so storing zeros instead."
                )

                print("\nThe following grid points are missing:")

                for i in range(flux_int.shape[0]):
                    if np.isnan(np.sum(flux_int[i, :])):
                        print("   - ", end="")
                        print(f"{parameters[0]} = {new_points[0][i]}, ", end="")
                        print(f"{parameters[1]} = {new_points[1][i]}, ", end="")
                        print(f"{parameters[2]} = {new_points[2][i]}, ", end="")
                        print(f"{parameters[3]} = {new_points[3][i]}")

        # ran_par_0 = np.random.randint(grid_shape[0], size=1000)
        # ran_par_1 = np.random.randint(grid_shape[1], size=1000)
        # ran_par_2 = np.random.randint(grid_shape[2], size=1000)
        # ran_par_3 = np.random.randint(grid_shape[3], size=1000)
        #
        # for z in range(ran_par_0.shape[0]):
        #     i = ran_par_0[z]
        #     j = ran_par_1[z]
        #     k = ran_par_2[z]
        #     m = ran_par_3[z]
        #
        #     if 0 < i < grid_shape[0]-1:
        #         check_low = np.isinf(np.sum(flux[i-1, j, k, m, ...]))
        #         check_up = np.isinf(np.sum(flux[i+1, j, k, m, ...]))
        #
        #         # Linear scaling of the intermediate Teff point
        #         scaling = (param_data[0][i] - param_data[0][i-1]) / \
        #                   (param_data[0][i+1] - param_data[0][i-1])
        #
        #         if not check_low and not check_up:
        #             flux_low = flux[i-1, j, k, m, ...]
        #             flux_up = flux[i+1, j, k, m, ...]
        #             flux[i, j, k, m, ...] = flux_low*(1.-scaling) + flux_up*scaling

    elif len(parameters) == 5:
        find_missing = np.zeros(grid_shape, dtype=bool)

        values = []
        points = [[], [], [], [], []]
        new_points = [[], [], [], [], []]

        for i in range(grid_shape[0]):
            for j in range(grid_shape[1]):
                for k in range(grid_shape[2]):
                    for m in range(grid_shape[3]):
                        for n in range(grid_shape[4]):
                            if np.isinf(np.sum(flux[i, j, k, m, n, ...])):
                                print("   - ", end="")
                                print(f"{parameters[0]} = {param_data[0][i]}, ", end="")
                                print(f"{parameters[1]} = {param_data[1][j]}, ", end="")
                                print(f"{parameters[2]} = {param_data[2][k]}, ", end="")
                                print(f"{parameters[3]} = {param_data[3][m]}, ", end="")
                                print(f"{parameters[4]} = {param_data[4][n]}")

                                if 0 < i < grid_shape[0] - 1:
                                    check_low = np.isinf(
                                        np.sum(flux[i - 1, j, k, m, n, ...])
                                    )
                                    check_up = np.isinf(
                                        np.sum(flux[i + 1, j, k, m, n, ...])
                                    )

                                    # Linear scaling of the intermediate Teff point
                                    scaling = (
                                        param_data[0][i] - param_data[0][i - 1]
                                    ) / (param_data[0][i + 1] - param_data[0][i - 1])

                                    if not check_low and not check_up:
                                        flux_low = flux[i - 1, j, k, m, n, ...]
                                        flux_up = flux[i + 1, j, k, m, n, ...]
                                        flux[i, j, k, m, n, ...] = (
                                            flux_low * (1.0 - scaling)
                                            + flux_up * scaling
                                        )
                                        count_interp += 1

                                    else:
                                        find_missing[i, j, k, m, n] = True

                                else:
                                    find_missing[i, j, k, m, n] = True

                            else:
                                points[0].append(param_data[0][i])
                                points[1].append(param_data[1][j])
                                points[2].append(param_data[2][k])
                                points[3].append(param_data[3][m])
                                points[4].append(param_data[4][n])

                                values.append(flux[i, j, k, m, n, ...])

                            new_points[0].append(param_data[0][i])
                            new_points[1].append(param_data[1][j])
                            new_points[2].append(param_data[2][k])
                            new_points[3].append(param_data[3][m])
                            new_points[4].append(param_data[4][n])

                            count_total += 1

        values = np.asarray(values)
        points = np.asarray(points)
        new_points = np.asarray(new_points)

        if np.sum(find_missing) > 0:
            flux_int = griddata(
                points.T, values, new_points.T, method="linear", fill_value=np.nan
            )

            count = 0

            for i in range(grid_shape[0]):
                for j in range(grid_shape[1]):
                    for k in range(grid_shape[2]):
                        for m in range(grid_shape[3]):
                            for n in range(grid_shape[4]):
                                if np.isnan(np.sum(flux_int[count, :])):
                                    count_missing += 1

                                elif np.isinf(np.sum(flux[i, j, k, m, n, ...])):
                                    flux[i, j, k, m, n, :] = flux_int[count, :]
                                    count_interp += 1

                                count += 1

            if count_missing > 0:
                warnings.warn(
                    f"Could not interpolate {count_missing} grid "
                    "points so storing zeros instead."
                )

                print("\nThe following grid points are missing:")

                for i in range(flux_int.shape[0]):
                    if np.isnan(np.sum(flux_int[i, :])):
                        print("   - ", end="")
                        print(f"{parameters[0]} = {new_points[0][i]}, ", end="")
                        print(f"{parameters[1]} = {new_points[1][i]}, ", end="")
                        print(f"{parameters[2]} = {new_points[2][i]}, ", end="")
                        print(f"{parameters[3]} = {new_points[3][i]}, ", end="")
                        print(f"{parameters[4]} = {new_points[4][i]}")

    else:
        raise ValueError(
            "The add_missing function is currently not compatible "
            "with more than 5 model parameters."
        )

    print(f"\nNumber of stored grid points: {count_total}")
    print(f"Number of interpolated grid points: {count_interp}")
    print(f"Number of missing grid points: {count_missing}")

    del database[f"models/{model}/flux"]
    database.create_dataset(f"models/{model}/flux", data=10.0**flux)


@typechecked
def correlation_to_covariance(
    cor_matrix: np.ndarray, spec_sigma: np.ndarray
) -> np.ndarray:
    """
    Parameters
    ----------
    cor_matrix : np.ndarray
        Correlation matrix of the spectrum.
    spec_sigma : np.ndarray
        Uncertainties (W m-2 um-1).

    Returns
    -------
    np.ndarray
        Covariance matrix of the spectrum.
    """

    cov_matrix = np.zeros(cor_matrix.shape)

    for i in range(cor_matrix.shape[0]):
        for j in range(cor_matrix.shape[1]):
            cov_matrix[i, j] = cor_matrix[i, j] * spec_sigma[i] * spec_sigma[j]

            if i == j:
                assert cor_matrix[i, j] == 1.0

    return cov_matrix


@typechecked
def convert_units(
    flux_in: np.ndarray, units_in: Tuple[str, str], convert_from: bool = True
) -> np.ndarray:
    """
    Function for converting the wavelength units to or from
    :math:`\\mu\\text{m}` and the flux units to or from
    :math:`\\text{W} \\text{m}^{-2} \\mu\\text{m}^{-1}`.

    Parameters
    ----------
    flux_in : np.ndarray
        Array with the input wavelengths and fluxes. The shape of the
        array should be (n_wavelengths, 3) or (n_wavelengths, 2) with
        the columns being the wavelengths, flux densities, and
        optionally the uncertainties. For photometric fluxes, the
        array should also be 2D but with a single row/wavelength.
    units_in : tuple(str, str)
        Tuple with the units of the wavelength ("um", "angstrom",
        "nm", "mm", "cm", "m", "Hz", "GHz") and the units of the
        flux density ("W m-2 um-1", "W m-2 m-1", "W m-2 Hz-1",
        "erg s-1 cm-2 angstrom-1" "erg s-1 cm-2 Hz-1", "uJy",
        mJy", "Jy", "MJy"). One can use "um" or "µm"
        interchangeably, and similarly "AA", "Å", "A", or
        "angstrom".
    convert_from : bool
        Convert from ``units_in`` to :math:`\\mu\\text{m}` and
        :math:`\\text{W} \\text{m}^{-2} \\mu\\text{m}^{-1}` when set to
        ``True``. Or, convert to ``units_in`` when set to ``False``.

    Returns
    -------
    np.ndarray
        Array with the output in the same shape as ``flux_in``.
    """

    speed_light = constants.LIGHT * 1e6  # (um s-1)

    flux_out = np.zeros(flux_in.shape)

    # Convert wavelengths to micrometer (um)

    wavel_units = ["um", "angstrom", "nm", "mm", "cm", "m", "Hz", "GHz"]

    if units_in[0] in ["um", "µm"]:
        flux_out[:, 0] = flux_in[:, 0].copy()

    elif units_in[0] in ["angstrom", "A", "AA", "Å"]:
        if convert_from:
            flux_out[:, 0] = flux_in[:, 0] * 1e-4
        else:
            flux_out[:, 0] = flux_in[:, 0] * 1e4

    elif units_in[0] == "nm":
        if convert_from:
            flux_out[:, 0] = flux_in[:, 0] * 1e-3
        else:
            flux_out[:, 0] = flux_in[:, 0] * 1e3

    elif units_in[0] == "mm":
        if convert_from:
            flux_out[:, 0] = flux_in[:, 0] * 1e3
        else:
            flux_out[:, 0] = flux_in[:, 0] * 1e-3

    elif units_in[0] == "cm":
        if convert_from:
            flux_out[:, 0] = flux_in[:, 0] * 1e4
        else:
            flux_out[:, 0] = flux_in[:, 0] * 1e-4

    elif units_in[0] == "m":
        if convert_from:
            flux_out[:, 0] = flux_in[:, 0] * 1e6
        else:
            flux_out[:, 0] = flux_in[:, 0] * 1e-6

    elif units_in[0] == "Hz":
        if convert_from:
            flux_out[:, 0] = speed_light / flux_in[:, 0]
        else:
            flux_out[:, 0] = speed_light / flux_in[:, 0]

    elif units_in[0] == "GHz":
        if convert_from:
            flux_out[:, 0] = speed_light / (1e9 * flux_in[:, 0])
        else:
            flux_out[:, 0] = 1e-9 * speed_light / flux_in[:, 0]

    else:
        raise ValueError(
            f"The wavelength units '{units_in[0]}' are not supported. "
            f"Please choose from the following units: {wavel_units}"
        )

    # Set wavelengths in micrometer

    if convert_from:
        wavel_micron = flux_out[:, 0].copy()
    else:
        wavel_micron = flux_in[:, 0].copy()

    # Convert flux density to W m-2 um-1

    flux_units = [
        "W m-2 um-1",
        "W m-2 m-1",
        "W m-2 Hz-1",
        "erg s-1 cm-2 angstrom-1",
        "erg s-1 cm-2 Hz-1",
        "uJy",
        "mJy",
        "Jy",
        "MJy",
    ]

    if units_in[1] in ["W m-2 um-1", "W m-2 µm-1"]:
        flux_out[:, 1] = flux_in[:, 1].copy()

        if flux_out.shape[1] == 3:
            flux_out[:, 2] = flux_in[:, 2].copy()

    elif units_in[1] == "W m-2 m-1":
        if convert_from:
            flux_out[:, 1] = flux_in[:, 1] * 1e-6
        else:
            flux_out[:, 1] = flux_in[:, 1] * 1e6

        if flux_out.shape[1] == 3:
            if convert_from:
                flux_out[:, 2] = flux_in[:, 2] * 1e-6
            else:
                flux_out[:, 2] = flux_in[:, 2] * 1e6

    elif units_in[1] == "W m-2 Hz-1":
        if convert_from:
            flux_out[:, 1] = flux_in[:, 1] * speed_light / wavel_micron**2
        else:
            flux_out[:, 1] = flux_in[:, 1] * wavel_micron**2 / speed_light

        if flux_out.shape[1] == 3:
            if convert_from:
                flux_out[:, 2] = flux_in[:, 2] * speed_light / wavel_micron**2
            else:
                flux_out[:, 2] = flux_in[:, 2] * wavel_micron**2 / speed_light

    elif units_in[1] == "erg s-1 cm-2 Hz-1":
        if convert_from:
            flux_out[:, 1] = flux_in[:, 1] * 1e-3 * speed_light / wavel_micron**2
        else:
            flux_out[:, 1] = flux_in[:, 1] * 1e3 * wavel_micron**2 / speed_light

        if flux_out.shape[1] == 3:
            if convert_from:
                flux_out[:, 2] = flux_in[:, 2] * 1e-3 * speed_light / wavel_micron**2
            else:
                flux_out[:, 2] = flux_in[:, 2] * 1e3 * wavel_micron**2 / speed_light

    elif units_in[1] in [
        "erg s-1 cm-2 angstrom-1",
        "erg s-1 cm-2 A-1",
        "erg s-1 cm-2 AA-1",
        "erg s-1 cm-2 Å-1",
    ]:
        if convert_from:
            flux_out[:, 1] = flux_in[:, 1] * 1e1
        else:
            flux_out[:, 1] = flux_in[:, 1] * 1e-1

        if flux_out.shape[1] == 3:
            if convert_from:
                flux_out[:, 2] = flux_in[:, 2] * 1e1
            else:
                flux_out[:, 2] = flux_in[:, 2] * 1e-1

    elif units_in[1] == "uJy":
        if convert_from:
            flux_out[:, 1] = flux_in[:, 1] * 1e-32 * speed_light / wavel_micron**2
        else:
            flux_out[:, 1] = flux_in[:, 1] * 1e32 * wavel_micron**2 / speed_light

        if flux_out.shape[1] == 3:
            if convert_from:
                flux_out[:, 2] = flux_in[:, 2] * 1e-32 * speed_light / wavel_micron**2
            else:
                flux_out[:, 2] = flux_in[:, 2] * 1e32 * wavel_micron**2 / speed_light

    elif units_in[1] == "mJy":
        if convert_from:
            flux_out[:, 1] = flux_in[:, 1] * 1e-29 * speed_light / wavel_micron**2
        else:
            flux_out[:, 1] = flux_in[:, 1] * 1e29 * wavel_micron**2 / speed_light

        if flux_out.shape[1] == 3:
            if convert_from:
                flux_out[:, 2] = flux_in[:, 2] * 1e-29 * speed_light / wavel_micron**2
            else:
                flux_out[:, 2] = flux_in[:, 2] * 1e29 * wavel_micron**2 / speed_light

    elif units_in[1] == "Jy":
        if convert_from:
            flux_out[:, 1] = flux_in[:, 1] * 1e-26 * speed_light / wavel_micron**2
        else:
            flux_out[:, 1] = flux_in[:, 1] * 1e26 * wavel_micron**2 / speed_light

        if flux_out.shape[1] == 3:
            if convert_from:
                flux_out[:, 2] = flux_in[:, 2] * 1e-26 * speed_light / wavel_micron**2
            else:
                flux_out[:, 2] = flux_in[:, 2] * 1e26 * wavel_micron**2 / speed_light

    elif units_in[1] == "MJy":
        if convert_from:
            flux_out[:, 1] = flux_in[:, 1] * 1e-20 * speed_light / wavel_micron**2
        else:
            flux_out[:, 1] = flux_in[:, 1] * 1e20 * wavel_micron**2 / speed_light

        if flux_out.shape[1] == 3:
            if convert_from:
                flux_out[:, 2] = flux_in[:, 2] * 1e-20 * speed_light / wavel_micron**2
            else:
                flux_out[:, 2] = flux_in[:, 2] * 1e20 * wavel_micron**2 / speed_light

    else:
        raise ValueError(
            f"The flux units '{units_in[1]}' are not supported. "
            f"Please choose from the following units: {flux_units}"
        )

    return flux_out


@typechecked
def remove_directory(dir_in: Path) -> None:
    """
    Function for removing all files and directories within a
    specified input directory.

    Parameters
    ----------
    dir_in : Path
        Directory that should be recursively removed, that is,
        including all the files and directories within ``dir_in``.

    Returns
    -------
    NoneType
        None
    """

    for path_item in dir_in.iterdir():
        if path_item.is_dir():
            remove_directory(path_item)
        else:
            path_item.unlink()

    dir_in.rmdir()



"""
Module for adding O5 through L3 SDSS stellar spectra from
Kesseli et al. (2017) to the database.
"""

from pathlib import Path

import h5py
import numpy as np
import pooch

from astropy.io import fits
from typeguard import typechecked

from species.util.data_util import extract_tarfile, remove_directory


@typechecked
def add_kesseli2017(input_path: str, database: h5py._hl.files.File) -> None:
    """
    Function for adding the SDSS stellar spectra from
    Kesseli et al. (2017) to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        The HDF5 database.

    Returns
    -------
    NoneType
        None
    """

    url = "https://cdsarc.u-strasbg.fr/viz-bin/nph-Cat/tar.gz?J/ApJS/230/16"
    input_file = "J_ApJS_230_16.tar.gz"
    data_file = Path(input_path) / input_file
    data_folder = Path(input_path) / "kesseli+2017/"

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash=None,
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    if data_folder.exists():
        remove_directory(data_folder)

    print(
        "\nUnpacking SDSS spectra from Kesseli et al. 2017 (145 MB)...",
        end="",
        flush=True,
    )
    extract_tarfile(str(data_file), str(data_folder))
    print(" [DONE]")

    fits_folder = Path(data_folder) / "fits"

    print_message = ""
    print()

    spec_files = sorted(fits_folder.glob("*"))

    for file_item in spec_files:
        data = fits.getdata(file_item, ext=1)

        wavelength = 1e-4 * 10.0 ** data["LogLam"]  # (um)
        flux = data["Flux"]  # Normalized units
        error = data["PropErr"]  # Normalized units

        name = file_item.stem.replace("_", " ")

        file_split = file_item.stem.split("_")
        sptype = file_split[0].split(".")[0]

        spdata = np.column_stack([wavelength, flux, error])

        empty_message = len(print_message) * " "
        print(f"\r{empty_message}", end="")

        print_message = f"Adding spectra... {name}"
        print(f"\r{print_message}", end="")

        dset = database.create_dataset(f"spectra/kesseli+2017/{name}", data=spdata)

        dset.attrs["name"] = str(name).encode()
        dset.attrs["sptype"] = str(sptype).encode()

    empty_message = len(print_message) * " "
    print(f"\r{empty_message}", end="")

    print_message = "Adding spectra... [DONE]"
    print(f"\r{print_message}")



from pathlib import Path
from zipfile import ZipFile

import h5py
import numpy as np
import pooch

from typeguard import typechecked

from species.core import constants


@typechecked
def add_sonora_diamondback(database: h5py._hl.files.File, input_path: str) -> None:
    """
    Function for adding the isochrone data of  `Sonora Diamondback
    <https://zenodo.org/records/12735103>`_ to the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    NoneType
        None
    """

    url = "https://zenodo.org/records/12735103/files/evolution.zip"

    input_file = "evolution.zip"
    data_folder = Path(input_path) / "sonora-diamondback-evolution"
    data_file = Path(data_folder) / input_file

    if not data_folder.exists():
        data_folder.mkdir()

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash="1efb4e5297060fa7b0329dec363e0bfff4f6132d2d11b655281cabea091d78ee",
            fname=input_file,
            path=data_folder,
            progressbar=True,
        )

    print("\nUnpacking Sonora Diamondback evolution (830 kB)...", end="", flush=True)
    with ZipFile(str(data_file)) as zip_object:
        zip_object.extractall(path=str(data_folder))
    print(" [DONE]")

    url = "https://zenodo.org/records/12735103/files/photometry.zip"

    input_file = "photometry.zip"
    data_file = Path(data_folder) / input_file

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash="7210f358c6da317d60a19dffb5b18a398e20a565aebaaa297e521ee9076bbc9c",
            fname=input_file,
            path=data_folder,
            progressbar=True,
        )

    print("\nUnpacking Sonora Diamondback photometry (239 kB)...", end="", flush=True)
    with ZipFile(str(data_file)) as zip_object:
        zip_object.extractall(path=str(data_folder))
    print(" [DONE]")

    iso_files = [
        "nc_m-0.5_age",
        "nc_m0.0_age",
        "nc_m+0.5_age",
        "hybrid_f2_m-0.5_age",
        "hybrid_f2_m0.0_age",
        "hybrid_f2_m+0.5_age",
        "hybrid-grav_f2_m-0.5_age",
        "hybrid-grav_f2_m0.0_age",
        "hybrid-grav_f2_m+0.5_age",
    ]

    labels = [
        "cloud-free, [M/H] = -0.5",
        "cloud-free, [M/H] = +0.0",
        "cloud-free, [M/H] = +0.5",
        "hybrid, fsed = 2, [M/H] = -0.5",
        "hybrid, fsed = 2, [M/H] = +0.0",
        "hybrid, fsed = 2, [M/H] = +0.5",
        "hybrid-grav, fsed = 2, [M/H] = -0.5",
        "hybrid-grav, fsed = 2, [M/H] = +0.0",
        "hybrid-grav, fsed = 2, [M/H] = +0.5",
    ]

    iso_tags = [
        "nc-0.5",
        "nc+0.0",
        "nc+0.5",
        "hybrid-0.5",
        "hybrid+0.0",
        "hybrid+0.5",
        "hybrid-grav-0.5",
        "hybrid-grav+0.0",
        "hybrid-grav+0.5",
    ]

    for iso_idx, iso_item in enumerate(iso_files):
        iso_file = f"evolution/{iso_item}"
        iso_path = Path(data_folder) / iso_file

        iso_data = []

        with open(str(iso_path), encoding="utf-8") as open_file:
            for line_idx, line_item in enumerate(open_file):
                if line_idx == 0 or "Gyr" in line_item:
                    continue

                # age(Gyr)  M/Msun  log(L/Lsun)  Teff(K)  log(g)  R/Rjup
                # The data files have R/Rsun, but should be R/Rjup
                param = list(filter(None, line_item.strip().split(" ")))
                param = list(map(float, param))

                param[0] = 1e3 * param[0]  # (Gyr) -> (Myr)

                param[1] = (
                    param[1] * constants.M_SUN / constants.M_JUP
                )  # (Msun) -> (Mjup)

                iso_data.append(
                    [param[0], param[1], param[2], param[3], param[4], param[5]]
                )

            print(
                f"\nAdding isochrones: Sonora {labels[iso_idx]}...", end="", flush=True
            )

            iso_data = np.array(iso_data)

            dset = database.create_dataset(
                f"isochrones/sonora-diamondback-{iso_tags[iso_idx]}/age",
                data=iso_data[:, 0],
            )  # (Myr)

            database.create_dataset(
                f"isochrones/sonora-diamondback-{iso_tags[iso_idx]}/mass",
                data=iso_data[:, 1],
            )  # (Mjup)

            database.create_dataset(
                f"isochrones/sonora-diamondback-{iso_tags[iso_idx]}/log_lum",
                data=iso_data[:, 2],
            )  # log(L/Lsun)

            database.create_dataset(
                f"isochrones/sonora-diamondback-{iso_tags[iso_idx]}/teff",
                data=iso_data[:, 3],
            )  # (K)

            database.create_dataset(
                f"isochrones/sonora-diamondback-{iso_tags[iso_idx]}/log_g",
                data=iso_data[:, 4],
            )  # log(g)

            database.create_dataset(
                f"isochrones/sonora-diamondback-{iso_tags[iso_idx]}/radius",
                data=iso_data[:, 5],
            )  # (Rjup)

            dset.attrs["model"] = "sonora-diamondback"

            print(" [DONE]")
            print(f"Database tag: sonora-diamondback-{iso_tags[iso_idx]}")



"""
Module with reading functionalities of color and magnitude
data from photometric and libraries.
"""

import os

from configparser import ConfigParser
from typing import Optional, Tuple

import h5py
import numpy as np

from typeguard import typechecked

from species.core.box import ColorColorBox, ColorMagBox, create_box
from species.read.read_spectrum import ReadSpectrum
from species.util.convert_util import apparent_to_absolute, parallax_to_distance
from species.util.core_util import print_section


class ReadColorMagnitude:
    """
    Class for reading color-magnitude data from the database.
    """

    @typechecked
    def __init__(
        self, library: str, filters_color: Tuple[str, str], filter_mag: str
    ) -> None:
        """
        Parameters
        ----------
        library : str
            Photometric ('vlm-plx' or 'leggett') or spectral ('irtf' or
            'spex') library.
        filters_color : tuple(str, str)
            Filter names for the color. For a photometric library,
            these have to be present in the database (typically in the
            MKO, 2MASS, or WISE system). For a spectral library, any
            filter names can be provided as long as they overlap with
            the wavelength range of the spectra.
        filter_mag : str
            Filter name for the absolute magnitudes (see also
            description of ``filters_color``).

        Returns
        -------
        NoneType
            None
        """

        print_section("Read color-magnitude")

        self.library = library
        self.filters_color = filters_color
        self.filter_mag = filter_mag

        if 'SPECIES_CONFIG' in os.environ:
            config_file = os.environ['SPECIES_CONFIG']
        else:
            config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]

        with h5py.File(self.database, "r") as hdf_file:
            if "photometry" in hdf_file and self.library in hdf_file["photometry"]:
                self.lib_type = "phot_lib"

            elif "spectra" in hdf_file and self.library in hdf_file["spectra"]:
                self.lib_type = "spec_lib"

            else:
                raise ValueError(
                    f"The '{self.library}' library is not present in the database."
                )

        print(f"Database tag: {self.library}")
        print(f"Library type: {self.lib_type}")
        print(f"Filters color: {self.filters_color}")
        print(f"Filter magnitude: {self.filter_mag}")

    @typechecked
    def get_color_magnitude(self, object_type: Optional[str] = None) -> ColorMagBox:
        """
        Function for extracting color-magnitude data from the selected
        library.

        Parameters
        ----------
        object_type : str, None
            Object type for which the colors and magnitudes are
            extracted. Either field dwarfs ('field') or
            young/low-gravity objects ('young'). All objects are
            selected if set to ``None``.

        Returns
        -------
        species.core.box.ColorMagBox
            Box with the colors and magnitudes.
        """

        print_section("Get color-magnitude")

        print(f"Object type: {object_type}")

        if self.lib_type == "phot_lib":
            with h5py.File(self.database, "r") as h5_file:
                sptype = np.asarray(h5_file[f"photometry/{self.library}/sptype"])
                parallax = np.asarray(h5_file[f"photometry/{self.library}/parallax"])
                parallax_error = np.asarray(
                    h5_file[f"photometry/{self.library}/parallax_error"]
                )
                flag = np.asarray(h5_file[f"photometry/{self.library}/flag"])
                obj_names = np.asarray(h5_file[f"photometry/{self.library}/name"])

                for i in range(sptype.shape[0]):
                    if isinstance(sptype[i], bytes):
                        sptype[i] = sptype[i].decode("utf-8")

                    if isinstance(flag[i], bytes):
                        flag[i] = flag[i].decode("utf-8")

                    if isinstance(obj_names[i], bytes):
                        obj_names[i] = obj_names[i].decode("utf-8")

            if object_type is None:
                indices = np.arange(0, np.size(sptype), 1)

            elif object_type == "field":
                indices = np.where(flag == "null")[0]

            elif object_type == "young":
                indices = []

                for j, object_flag in enumerate(flag):
                    if "young" in object_flag:
                        indices.append(j)

                    elif "lowg" in object_flag:
                        indices.append(j)

                indices = np.array(indices)

            if indices.size > 0:
                with h5py.File(self.database, "r") as h5_file:
                    filter_list = []

                    for item in list(h5_file[f"photometry/{self.library}/MKO"]):
                        filter_list.append(f"MKO/{item}")

                    for item in list(h5_file[f"photometry/{self.library}/2MASS"]):
                        filter_list.append(f"2MASS/{item}")

                    for item in list(h5_file[f"photometry/{self.library}/Keck"]):
                        filter_list.append(f"Keck/{item}")

                    if f"photometry/{self.library}/{self.filters_color[0]}" in h5_file:
                        mag1 = np.asarray(
                            h5_file[
                                f"photometry/{self.library}/{self.filters_color[0]}"
                            ]
                        )

                    else:
                        raise ValueError(
                            f"The '{self.filters_color[0]}' "
                            "filter is not available in the "
                            f"'{self.library}' library. Please "
                            "select any of the following "
                            f"filters: {filter_list}"
                        )

                    if f"photometry/{self.library}/{self.filters_color[1]}" in h5_file:
                        mag2 = np.asarray(
                            h5_file[
                                f"photometry/{self.library}/{self.filters_color[1]}"
                            ]
                        )

                    else:
                        raise ValueError(
                            f"The '{self.filters_color[1]}' "
                            "filter is not available in the "
                            f"'{self.library}' library. Please "
                            "select any of the following "
                            f"filters: {filter_list}"
                        )

            else:
                raise ValueError(
                    f"There is not data available from '{self.library}' for "
                    f"'{object_type}' type objects with the chosen filters."
                )

            color = mag1 - mag2

            distance = parallax_to_distance((parallax, parallax_error))

            if self.filter_mag == self.filters_color[0]:
                mag, _ = apparent_to_absolute((mag1, None), (distance[0], distance[1]))

            elif self.filter_mag == self.filters_color[1]:
                mag, _ = apparent_to_absolute((mag2, None), (distance[0], distance[1]))

            else:
                raise ValueError(
                    "The argument of filter_mag is set "
                    f"to {self.filter_mag} but should be "
                    "the same as one of the two values of "
                    "filters_color."
                )

            color = color[indices]
            mag = mag[indices]
            sptype = sptype[indices]
            obj_names = obj_names[indices]

            indices = []
            for i in range(color.size):
                if not np.isnan(color[i]) and not np.isnan(mag[i]):
                    indices.append(i)

            colormag_box = create_box(
                boxtype="colormag",
                library=self.library,
                object_type=object_type,
                filters_color=self.filters_color,
                filter_mag=self.filter_mag,
                color=color[indices],
                magnitude=mag[indices],
                sptype=sptype[indices],
                names=obj_names[indices],
            )

        elif self.lib_type == "spec_lib":
            read_spec_0 = ReadSpectrum(
                spec_library=self.library, filter_name=self.filters_color[0]
            )

            read_spec_1 = ReadSpectrum(
                spec_library=self.library, filter_name=self.filters_color[1]
            )

            read_spec_2 = ReadSpectrum(
                spec_library=self.library, filter_name=self.filter_mag
            )

            phot_box_0 = read_spec_0.get_magnitude(sptypes=None)
            phot_box_1 = read_spec_1.get_magnitude(sptypes=None)
            phot_box_2 = read_spec_2.get_magnitude(sptypes=None)

            colormag_box = create_box(
                boxtype="colormag",
                library=self.library,
                object_type=object_type,
                filters_color=self.filters_color,
                filter_mag=self.filter_mag,
                color=phot_box_0.app_mag[:, 0] - phot_box_1.app_mag[:, 0],
                magnitude=phot_box_2.abs_mag[:, 0],
                sptype=phot_box_0.sptype,
                names=None,
            )

        n_objects = len(colormag_box.sptype)
        print(f"Returning ColorMagBox with {n_objects} objects")

        return colormag_box


class ReadColorColor:
    """
    Class for reading color-color data from the database.
    """

    @typechecked
    def __init__(
        self, library: str, filters_colors: Tuple[Tuple[str, str], Tuple[str, str]]
    ) -> None:
        """
        Parameters
        ----------
        library : str
            Photometric ('vlm-plx' or 'leggett') or spectral ('irtf' or
            'spex') library.
        filters_colors : tuple(tuple(str, str), tuple(str, str))
            Filter names for the colors. For a photometric library,
            these have to be present in the database (typically in the
            MKO, 2MASS, or WISE system). For a spectral library, any
            filter names can be provided as long as they overlap with
            the wavelength range of the spectra.

        Returns
        -------
        NoneType
            None
        """

        print_section("Read color-magnitude")

        self.library = library
        self.filters_colors = filters_colors

        if 'SPECIES_CONFIG' in os.environ:
            config_file = os.environ['SPECIES_CONFIG']
        else:
            config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]

        with h5py.File(self.database, "r") as hdf_file:
            if "photometry" in hdf_file and self.library in hdf_file["photometry"]:
                self.lib_type = "phot_lib"

            elif "spectra" in hdf_file and self.library in hdf_file["spectra"]:
                self.lib_type = "spec_lib"

            else:
                raise ValueError(
                    f"The '{self.library}' library is not present in the database."
                )

        print(f"Database tag: {self.library}")
        print(f"Library type: {self.lib_type}")
        print(f"Filters colors: {self.filters_colors}")

    @typechecked
    def get_color_color(self, object_type: Optional[str] = None) -> ColorColorBox:
        """
        Function for extracting color-color data from the selected
        library.

        Parameters
        ----------
        object_type : str, None
            Object type for which the colors and magnitudes are
            extracted. Either field dwarfs ('field') or
            young/low-gravity objects ('young'). All objects are
            selected if set to ``None``.

        Returns
        -------
        species.core.box.ColorColorBox
            Box with the colors.
        """

        print_section("Get color-magnitude")

        print(f"Object type: {object_type}")

        if self.lib_type == "phot_lib":
            h5_file = h5py.File(self.database, "r")

            sptype = np.asarray(h5_file[f"photometry/{self.library}/sptype"])
            flag = np.asarray(h5_file[f"photometry/{self.library}/flag"])
            obj_names = np.asarray(h5_file[f"photometry/{self.library}/name"])

            if object_type is None:
                indices = np.arange(0, np.size(sptype), 1)

            elif object_type == "field":
                indices = np.where(flag == "null")[0]

            elif object_type == "young":
                indices = []

                for j, object_flag in enumerate(flag):
                    if "young" in object_flag:
                        indices.append(j)

                    elif "lowg" in object_flag:
                        indices.append(j)

                indices = np.array(indices)

            mag1 = np.asarray(
                h5_file[f"photometry/{self.library}/{self.filters_colors[0][0]}"]
            )
            mag2 = np.asarray(
                h5_file[f"photometry/{self.library}/{self.filters_colors[0][1]}"]
            )
            mag3 = np.asarray(
                h5_file[f"photometry/{self.library}/{self.filters_colors[1][0]}"]
            )
            mag4 = np.asarray(
                h5_file[f"photometry/{self.library}/{self.filters_colors[1][1]}"]
            )

            color1 = mag1 - mag2
            color2 = mag3 - mag4

            color1 = color1[indices]
            color2 = color2[indices]
            sptype = sptype[indices]
            obj_names = obj_names[indices]

            indices = []
            for i in range(color1.size):
                if not np.isnan(color1[i]) and not np.isnan(color2[i]):
                    indices.append(i)

            colorbox = create_box(
                boxtype="colorcolor",
                library=self.library,
                object_type=object_type,
                filters=self.filters_colors,
                color1=color1[indices],
                color2=color2[indices],
                sptype=sptype[indices],
                names=obj_names[indices],
            )

            h5_file.close()

        elif self.lib_type == "spec_lib":
            read_spec_0 = ReadSpectrum(
                spec_library=self.library, filter_name=self.filters_colors[0][0]
            )

            read_spec_1 = ReadSpectrum(
                spec_library=self.library, filter_name=self.filters_colors[0][1]
            )

            read_spec_2 = ReadSpectrum(
                spec_library=self.library, filter_name=self.filters_colors[1][0]
            )

            read_spec_3 = ReadSpectrum(
                spec_library=self.library, filter_name=self.filters_colors[1][1]
            )

            phot_box_0 = read_spec_0.get_magnitude(sptypes=None)
            phot_box_1 = read_spec_1.get_magnitude(sptypes=None)
            phot_box_2 = read_spec_2.get_magnitude(sptypes=None)
            phot_box_3 = read_spec_3.get_magnitude(sptypes=None)

            colorbox = create_box(
                boxtype="colorcolor",
                library=self.library,
                object_type=object_type,
                filters=self.filters_colors,
                color1=phot_box_0.app_mag[:, 0] - phot_box_1.app_mag[:, 0],
                color2=phot_box_2.app_mag[:, 0] - phot_box_3.app_mag[:, 0],
                sptype=phot_box_0.sptype,
                names=None,
            )

        n_objects = len(colorbox.sptype)
        print(f"Returning ColorColorBox with {n_objects} objects")

        return colorbox



from pathlib import Path

import h5py
import pooch

from typeguard import typechecked

from species.data.isochrone_data.iso_manual import add_manual


@typechecked
def add_ames(database: h5py._hl.files.File, input_path: str) -> None:
    """
    Function for adding the AMES-Cond and AMES-Dusty
    isochrone data to the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    NoneType
        None
    """

    url_list = [
        "https://home.strw.leidenuniv.nl/~stolker/species/"
        "model.AMES-Cond-2000.M-0.0.MKO.Vega",
        "https://home.strw.leidenuniv.nl/~stolker/species/"
        "model.AMES-dusty.M-0.0.MKO.Vega",
    ]

    file_hash = [
        "fc04e6f7c02982bb3187b55cdefc2464e3f1564fb8026a8958967cb889f0f581",
        "c7ba32ae10111c9ca692bf75154edac70b050c06cae211b421e1473725d6380c",
    ]

    iso_tags = ["ames-cond", "ames-dusty"]

    for url_idx, url_item in enumerate(url_list):
        input_file = url_item.split("/")[-1]
        data_file = Path(input_path) / input_file

        if not data_file.exists():
            print()

            pooch.retrieve(
                url=url_item,
                known_hash=file_hash[url_idx],
                fname=input_file,
                path=input_path,
                progressbar=True,
            )

        add_manual(
            database=database,
            tag=iso_tags[url_idx],
            file_name=str(data_file),
            model_name="ames",
        )
